import numpy as np
import astropy.units as u
from scipy.interpolate import interp1d # numpy should be faster...
import scipy.special as sps
#import warnings

def log_profile(args,r_bins,radii,alphas=[],rintmax=[],finite=False):
    """   
    Performs logarithmic interpolation and extrapolation. This can easily be improved, but hasn't been the bottleneck.
    
    r_bins and radii must be in the same units.
    Parameters
    ----------
    args : class:`numpy.ndarray`
       binned values (assumed to be pressure)
    r_bins : class:`numpy.ndarray`
       bin edges
    radii : class:`numpy.ndarray`
       radial points onto which you will interpolate or extrapolate
    alphas: array-like, optional
       If zeros or an empty list, this function will calculate power law index between bins.
    rintmax : array-like, optional
       If zeros or an empty list, integration extends to infinity. Default is [].
    finite : bool
       Do not integrate to infinity; stop at last value in r_bins.

    Returns
    -------
    presprof : Interpolated (pressure) profile
    alphas : The power-law indices between bins (e.g. pressures)
    """

    #warnings.showwarning = handle_warning
    
    #r_uniqe = np.unique(r_bins)
    #mybins=[0] + r_bins
    if not finite:
        mybins=np.insert(r_bins,0,0)
        mybins[-1]=-1
        #import pdb;pdb.set_trace()
    else:
        mybins = r_bins.copy()
        #import pdb;pdb.set_trace()
        
    presprof=np.zeros(len(radii))
    #        ifht=('alphas' in dir())
    ifht=('alphas' in locals())
    if ifht: 
        if sum(alphas) != 0:
            aset=1
        else:
            aset=0
            alphas = np.zeros(len(r_bins))
    else:
        aset=0
        alphas = np.zeros(len(r_bins))
        
    ycyl=0.0 # Y_500,cyl integration.
    ### Actually, I am not calculating Ycyl here!!! Stupid me. I could 
    ### calculate Ysph here though...

    badind = 0
    nfor = len(r_bins)-1 if finite else len(r_bins)
    for idx in range(nfor):
        rin=mybins[idx]
        if idx+1 == len(mybins):
            import pdb;pdb.set_trace()
        rout=mybins[idx+1]
        epsnot=args[idx]
        alpha=alphas[idx]
        if rin == rout:
            badind = idx
            continue        
        if rin == 0:
            lr=np.log10(mybins[idx+2]/mybins[idx+1])
            lp=np.log10(args[idx+1]/args[idx])
            if aset == 0:
                alpha=-lp/lr
                #                myind=np.where((radii < rout) & (radii >= rin))
            myind=(radii < rout) & (radii >= rin)
            myrad=radii[myind]
            mypres=epsnot*(myrad/rout)**(-alpha)
            # I'm often not using the integrated y; but we can keep it for now.
            yint = 0 if rin == 0 else 1.0 - (rin/rout)**(2-alpha)
            # but this ensures an error if alpha >2 ...
            rnot=rout
        elif rout == -1:
            lr=np.log10(r_bins[idx]/r_bins[idx-1])
            lp=np.log10(args[idx]/args[idx-1])
            if aset == 0:
                alpha=-lp/lr
            epsnot=args[idx-1]
            #                myind=(radii >= rin)
            myind=np.where(radii >= rin)
            myrad=radii[myind]
            mypres=epsnot*(myrad/rin)**(-alpha)
            rnot=rin
            yint = -1.0
            if np.sum(rintmax) > 0:
                yint = (rintmax/rnot)**(2-alpha) - 1.0

        else:
            lr=np.log10(mybins[idx+1]/mybins[idx])
            lp=np.log10(args[idx]/args[idx-1])
            if aset == 0:
                alpha=-lp/lr
            myind=np.where((radii < rout) & (radii >= rin))
            #                myind=(radii < rout) & (radii >= rin)
            myrad=radii[myind]
            mypres=epsnot*(myrad/rout)**(-alpha)
            rnot=rin
            yint = (rout/rin)**(2-alpha) - 1.0
            
        presprof[myind]=mypres
        if aset == 0:
            alphas[idx]=alpha

        ypref = 2*np.pi*epsnot*(rnot**2)/(2-alpha)
        if np.sum(rintmax) > 0:
            if rin < rintmax:
                if (rout > 0) & (rout <= rintmax):
                    yint=(rintmax/rnot)**(2-alpha)-1.0
                ycyl=ycyl + ypref*yint

            return presprof,alphas,ycyl
        # back to this placent

    if badind > 0:
        alphas = np.delete(alphas,badind)
        
    if aset == 0:
        return presprof,alphas
    else:
        return presprof

def binsky(args,r_bins,theta_range,theta,inalphas=[]):
    """
    Returns a surface brightness map for a binned profile, slopes, and radial integrals.
    
    Parameters
    ----------
     args : class:`numpy.ndarray`
       Pressure for each bin used
    r_bins: class:`numpy.ndarray`
       bin edges
    theta_range : class:`numpy.ndarray`
       Highly sampled array of radii.
    theta : class:`numpy.ndarray`
       A map of azimuthal angles.
    inalphas : list
       Generally best to leave as an empty list.

    Returns
    -------
    outmap: class:`numpy.ndarray`
    alphas: class:`numpy.ndarray`
    integrals: class:`numpy.ndarray`
    """
    Int_Pres,alphas,integrals = analytic_shells(r_bins,args,theta_range,alphas=inalphas)
    fint = interp1d(theta_range, Int_Pres, bounds_error = False, 
                    fill_value = 0)
    nx, ny = theta.shape
    #map = np.float64(fint(theta.reshape(nx * ny))) # Type 17 = float? (Implicitly float 32?)
    map = fint(theta.reshape(nx * ny)) # Type 17 = float? (Implicitly float 32?)
    outmap = map.reshape(nx,ny)

    return outmap,alphas,integrals

def prep_SZ_binsky(pressure, temp_iso, geoparams=None):
    """
    Small function, intended to do more (for relativistic corrections), but currently only allows for one ICM temperature.

    Parameters
    ----------
    pressure : array-like
        array of electron pressures
    temp_iso : float
       isothermal temperature
    geoparams : array-like
        [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]

    Returns
    -------
    edensity: class:`numpy.ndarray`
       Proxy for electron density (per cubic cm)
    etemperature: class:`numpy.ndarray`
       Proxy for electron temperature (times Bolzmann Constant; keV)
    geoparams: array-like
       [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    """
    edensity = np.array(pressure) / temp_iso
    etemperature = np.array(pressure)*0 + temp_iso
    if geoparams == None:
        geoparams = [0,0,0,1,1,1,0,0] # Spherical Geometry

    return edensity, etemperature, geoparams

def integrate_profiles(epressure, geoparams,r_bins,theta_range,inalphas=[],
                       beta=0.0,betaz=None,finint=False,narm=False,fixalpha=False,strad=False,
                       array="2",fullSZcorr=False,SZtot=False,columnDen=False,Comptony=True,
                       instrument='MUSTANG2',negvals=None,tmax=0):
    """
    Returns a Compton y profile for an input pressure profile.
    
    Parameters
    ----------
    epressure :  class:`numpy.ndarray`
       Electron pressure in units such that its integral over theta_range (itself in radians), results in the unitless Compton y parameter.
    geoparams : array-like
       [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    r_bins : array-like
       The (elliptical) bins, in radians, for the profile. 
    theta_range : class:`numpy.ndarray` 
       The range of angles for which to create a 1D profile (which can then be interpolated)
    inalphas :  array-like, optional
       This should generally be an empty list, unless modelling shocks.
    beta :float
       Fraction of the speed of light of the cluster bulk (peculiar) motion.
    betaz : float
       Fraction of the speed of light of the cluster along the line of sight.
    finint : bool
       Integrate out to last finite (defined) bin. Defaults to False
    narm : bool
       Normalized at R_Min. Defaults to False
    fixalpha : bool
       Fix alpha (to whatever inalpha is); useful for shocks.
    strad : bool
       STrict RADii; if the pressure model has to obey strict placements of radii, use this!
    array :str
       only used with NIKA2 data; which detector array is being used?
    fullSZcorr : bool
       integrate relativistic corrections along line of sight?
    SZtot : bool
       total SZ signal... not really useful
    columnDen : bool
       Set to true if you want to return the column density...?
    Comptony : bool
       When set (by default), returns Comptony profile
    instrument : str
       MUSTANG-2 by default. Used in relativistic calculations.
    negvals : class:`numpy.ndarray`(dtype=bool)
       Boolean array, the length of density_proxy.

    Returns
    -------
    Int_Pres : class:`numpy.ndarray`
       The combined los integrals over theta_range
    alphas : class:`numpy.ndarray`
       The power-law indices between bins (e.g. pressures)
    integrals : class:`numpy.ndarray`
       The los integral per bin (e.g. per shell), each across theta_range

    Notes
    __________
    * Ella should be set to 1. Therefore, define Ellb relative to Ella (and likewise with Ellc)
    * Xi is a sneaky gem.
    * integrals allows for sneakiness.
    """
    if betaz == None:
        betaz = beta
### If geoparams[6] > 0, then we are modelling some non-ellipsoid...perhaps a shock. If the opening angle
### is not set, then this will create a bimodal (bipolar) model component, which we almost certainly don't
### want. If we do want a bimodal component, then I think a better override is to use geoparams[7]= 2 pi.

    #eff_pres = np.zeros(len(etemperature)); y_press= np.zeros(len(etemperature))
                       
    Int_Pres,alphas,integrals = analytic_shells(r_bins,epressure,theta_range,alphas=inalphas,shockxi=geoparams[6],
                                                finite=finint,narm=narm,fixalpha=fixalpha,strad=strad,
                                                negvals=negvals,tmax=tmax)
    
    return Int_Pres,alphas,integrals
        
def general_gridding(xymap,theta_range,r_bins,geoparams,finite=False,taper='normal',
                     integrals=0,Int_Pres=0,ell_int=0,tap_int=0,oldvs=False,xyinas=True):
    """
    Returns a surface brightness map for a binned los-integrated profile (Int_Pres).
    
    Parameters
    ----------
    xymap: tuple of class:`numpy.ndarray`
       A tuple (x,y) where x and y are grids of their respective coordinates in << arceconds >>
    theta_range: class:`numpy.ndarray`
       Array of radii for the corresponding Int_Pres
    geoparams:array-like
       [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]

    Notes
    __________
    * Ella should be set to 1. Therefore, define Ellb relative to Ella (and likewise with Ellc)
    
    Returns
    -------
    mymap : class:`numpy.ndarray`
       A 2D map from input radial surface brightness profile
    """
    
    if geoparams[6] > 0.0:
        x,y = xymap;  mymap = np.zeros(x.shape); myrs = r_bins
        if geoparams[7] == 0:
            geoparams[7] = np.pi 
        if finite == True:
            myrs = myrs[:-1]
        #for idx, val in enumerate(myrs):
        #    if val == 0: val=r_bins[idx+1] # Correct units? I think so.
        #    if taper == 'inverse':
        #        mymap += grid_profile(theta_range, ell_int[idx,:], xymap, geoparams=geoparams)
        #        mymap -= grid_profile(theta_range, tap_int[idx,:], xymap, geoparams=geoparams,myscale=val,axis='y')
        #    else:
        #        mymap += grid_profile(theta_range, integrals[idx,:], xymap, geoparams=geoparams,myscale=val,axis='x')
        ######################################################################################
        ### The following has been rewritten 30 Mar 2018, in hopes of being faster.
        if myrs[0] == 0: myrs[0]=myrs[1] # Correct units? I think so.
        if taper == 'inverse':
            for my_int_add, my_int_sub, val in zip(ell_int, tap_int,myrs):
                mymap += grid_profile(theta_range, my_int_add, xymap, geoparams=geoparams,xyinas=xyinas)
                mymap -= grid_profile(theta_range, my_int_sub, xymap, geoparams=geoparams,myscale=val,axis='y',xyinas=xyinas)
        else:
            if oldvs == True:
                for my_int, val in zip(integrals, myrs):
                    mymap += grid_profile(theta_range, my_int, xymap, geoparams=geoparams,myscale=val,axis='x',xyinas=xyinas)
            else:
                mymap=iter_grid_profile_v2(integrals, myrs, theta_range, xymap, geoparams=geoparams,axis='x',xyinas=xyinas)
         ######################################################################################
    else:  
        mymap = grid_profile(theta_range, Int_Pres, xymap, geoparams=geoparams,myscale=1.0,xyinas=xyinas)

    return mymap


#########################################################################################################
### + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + ###
###+ + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + +###
### + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + ###
###+ + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + +###
### + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + ###
###+ + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + +###
### + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + ###
###                                                                                                   ###
###                         Let's try to do things in a more general way                              ###
###                                                                                                   ###
###+ + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + +###
### + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + ###
###+ + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + +###
### + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + ###
###+ + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + +###
### + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + ###
###+ + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + +###
#########################################################################################################


def binsky_SZ_general(epressure, geoparams,r_bins,theta_range,xymap,
                      inalphas=[],beta=0.0,betaz=None,finite=False,narm=False,fixalpha=False,
                      strad=False,array="2",instrument='MUSTANG2',taper='normal'):
    """
    Returns a surface brightness map for a binned profile fit, with far more generality than previously done.
    
    Parameters
    ----------
    epressure : array-like
       The electron pressure (no units in Python, but otherwise should be in cm**-3 keV**-1)
    geoparams : array-like
       [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    r_bins : array-like
       The (elliptical) bins for the profile. 
    theta_range : class:`numpy.ndarray`
       Array of radii for the corresponding Int_Pres
    xymap : tuple of class:`numpy.ndarray`
       A tuple (x,y) where x and y are grids of their respective coordinates in << arceconds >>
    inalphas : array-like
       Nothing to see here. Move along.
    beta : float
       Fraction of the speed of light of the cluster bulk (peculiar) motion.
    betaz : float
       Fraction of the speed of light of the cluster along the line of sight.
    finite : bool
       Integrate out to last finite (defined) bin.
    narm : bool
       Normalized at R_Min. This is important for integrating shells.
    strad : bool
       STrict RADii; if the pressure model has to obey strict placements of radii, use this!

    Notes
    __________
    * Ella should be set to 1. Therefore, define Ellb relative to Ella (and likewise with Ellc)
    
    Returns
    -------
    mymap : class:`numpy.ndarray`
       A 2D map from input radial surface brightness profile   
    """
    if betaz == None:
        betaz = beta
### If geoparams[6] > 0, then we are modelling some non-ellipsoid...perhaps a shock. If the opening angle
### is not set, then this will create a bimodal (bipolar) model component, which we almost certainly don't
### want. If we do want a bimodal component, then I think a better override is to use geoparams[7]= 2 pi.

    if geoparams[6] > 0:
        if geoparams[7] == 0:
            geoparams[7] = np.pi 
    
    map,alphas,integrals = binsky_general(epressure,geoparams,r_bins,theta_range,xymap,inalphas=inalphas,
                                          finite=finite,narm=narm,taper=taper,fixalpha=fixalpha,strad=strad)

    return map,alphas,integrals
        
def binsky_general(vals,geoparams,r_bins,theta_range,xymap,inalphas=[],
                   finite=False,narm=False,taper='normal',fixalpha=False,strad=False):
    """
    Returns a surface brightness map for a binned profile fit 
    
    Parameters
    __________
    vals : array-like
       The electron pressure (no units in Python, but otherwise should be in cm**-3 keV**-1)
    geoparams : array-like
       [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    r_bins : array-like
       The (elliptical) bins for the profile. 
    theta_range  : array-like
       The range of angles for which to create a 1D profile (which can then be interpolated)
    xymap : tuple
       A tuple (x,y) where x and y are grids of their respective coordinates in << arceconds >>
    inalphas : array-like
       Nothing to see here. Move along.
    beta : float
       Fraction of the speed of light of the cluster bulk (peculiar) motion.
    betaz : float
       Fraction of the speed of light of the cluster along the line of sight.
    finite : bool
       Integrate out to last finite (defined) bin.
    narm : bool
       Normalized at R_Min. This is important for integrating shells.
    strad : bool
       STrict RADii; if the pressure model has to obey strict placements of radii, use this!
    
    Notes:
    __________
    --> We should consider Ella to be RESTRICTED to 1. That is, Ellb and Ellc should always be calculated
    relative to the x-axis parameter.
    
    Returns
    -------
    mymap : class:`numpy.ndarray`
       A map that accounts for a range of geometrical restrictions. The integrals may not be applicable.

    """

    if taper == 'inverse':
        Ell_Pres,alphas,ell_int = analytic_shells(r_bins,vals,theta_range,alphas=inalphas,finite=finite,
                                                  narm=narm,fixalpha=fixalpha,strad=strad)
        Tap_Pres,tap_alph,tap_int = analytic_shells(r_bins,vals,theta_range,alphas=inalphas,
                                                    shockxi=geoparams[6],finite=finite,narm=narm,
                                                    fixalpha=fixalpha,strad=strad)
        integrals = ell_int - tap_int
    else:
        Int_Pres,alphas,integrals = analytic_shells(r_bins,vals,theta_range,alphas=inalphas,
                                                    shockxi=geoparams[6],finite=finite,narm=narm,
                                                    fixalpha=fixalpha,strad=strad)

############################################################################

    map = general_gridding(xymap,r_bins,geoparams,finite,narm,taper,strad,
                           integrals,Int_Pres,ell_int,tap_int)

    return map,alphas,integrals

def grid_profile(theta_range, profile, xymap, geoparams=[0,0,0,1,1,1,0,0],myscale=1.0,axis='z',
                 xyinas=True):
    """   
    Grids a sufficiently fine-resolution profile.

    Parameters
    __________
    theta_range: class:`numpy.ndarray`
       Abscisca (radii, in radians) for profile
    profile: class:`numpy.ndarray`
       Ordinate profile values
    xymap: tuple
       A tuple (x,y) where x and y are grids of their respective coordinates in << arceconds >>
    geoparams: array-like
       [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    myscale: float
       Rescale the profile (e.g. to account for elongation along the line of sight)
    axis: str
       scale about x, y, or z
    xyinas: bool
       xymap is in arcseconds (yes)

    Returns
    -------
    mymap : class:`numpy.ndarray`
       A map that accounts for a range of geometrical restrictions.
    """
    
    ### Get new grid:
    (x,y) = xymap
    x,y = rot_trans_grid(x,y,geoparams[0],geoparams[1],geoparams[2])
    x,y = get_ell_rads(x,y,geoparams[3],geoparams[4])
    radmap = np.sqrt(x**2 + y**2)
    theta = radmap*(u.arcsec).to("radian");  theta_min = np.min(theta_range)
#    import pdb;pdb.set_trace()
    bi=np.where(theta < theta_min);   theta[bi]=theta_min

    nx, ny = theta.shape
    fint = interp1d(theta_range, profile, bounds_error = False, fill_value = 0)
    #mymap = np.float64(fint(theta.reshape(nx * ny))) # Type 17 = float? (Implicitly float 32?)
    mymap = fint(theta.reshape(nx * ny)) # Type 17 = float? (Implicitly float 32?)
    mymap = mymap.reshape(nx,ny)
    ### And a couple more *necessary* modification:
    ### Where we want to scale it by a certain r_bin, given in radians. We also want to scale by "Ella", if axis='x':

    a2r   = (u.arcsec).to("radian")
    conv = a2r if xyinas else 1.0  # Is xymap in arcseconds or radions?
    
    if axis == 'x':
        xell = (x/(geoparams[3]*myscale))*conv # x is initially presented in arcseconds
        modmap = geoparams[5]*(xell**2)**(geoparams[6]) # Consistent with model creation??? (26 July 2017)
    if axis == 'y':
        yell = (y/(geoparams[4]*myscale))*conv # x is initially presented in arcseconds
        modmap = geoparams[5]*(yell**2)**(geoparams[6]) # Consistent with model creation??? (26 July 2017)
    if axis == 'z':
        modmap = mymap*0.0 + geoparams[5]      # Just the plain old LOS elongation factor
    
    mymap = mymap*modmap   # Very important to be precise here.
#    import pdb;pdb.set_trace()
    if geoparams[7] > 0:
#        angmap = np.arctan2(x,y)
        angmap = np.arctan2(y,x)
        #        gi = np.where(abs(angmap) < geoparams[7]/2.0)
        bi = np.where(abs(angmap) > geoparams[7]/2.0)
        #        import pdb; pdb.set_trace()
        mymap[bi] = 0.0

    return mymap

def get_ell_rads(x,y,ella,ellb):
    """   
    Get ellipsoidal radii from x,y standard

    Parameters
    __________
    :param x: coordinate along major axis (a) 
    :type x: class:`numpy.ndarray`
    :param y: coordinate along minor axis (b) 
    :type y: class:`numpy.ndarray`
    :param ella: scaling along major axis (should stay 1)
    :type ella: float
    :param ellb: scaling along minor axis
    :type ella: float

    """
    
    xnew = x/ella ; ynew = y/ellb

    return xnew, ynew
    
def rot_trans_grid(x,y,xs,ys,rot_rad):
    """   
    Shift and rotate coordinates

    Parameters
    __________
    x: class:`numpy.ndarray`
       coordinate along major axis (a) 
    :param y: coordinate along minor axis (b) 
    :type y: class:`numpy.ndarray`
    :param xs: translation along x-axis
    :type xs: float
    :param ys: translation along y-axis
    :type ys: float
    :param rot_rad: rotation angle, in radians
    :type rot_rad: float

    """

    xnew = (x - xs)*np.cos(rot_rad) + (y - ys)*np.sin(rot_rad)
    ynew = (y - ys)*np.cos(rot_rad) - (x - xs)*np.sin(rot_rad)

    return xnew,ynew

def ycylfromprof(Int_Pres,theta_range,theta_max):
    """
    Integrate Int_Pres over area to get y_cyl

    Parameters
    __________
    :param Int_Pres: array of Compton y values
    :type Int_Pres: class:`numpy.ndarray`
    :param theta_range: array of radii, in radians
    :type theta_range: class:`numpy.ndarray`
    :param theta_max: perform integral within this radius.
    :type theta_max: float

    """
    i=1 # Start at the second entry!
    Ycyl=0
    while i < len(theta_range):
        if theta_range[i] < theta_max:
            dtheta = theta_range[i]-theta_range[i-1]
            Yshell = 2.0*np.pi*theta_range[i]*dtheta*Int_Pres[i]
            Ycyl = Ycyl + Yshell
        i+=1

    return Ycyl

def analytic_shells(r_bins,vals,theta,alphas=[],shockxi=0.0,fixalpha=False,
                    finite=False,narm=False,strad=False,negvals=None,tmax=0):
    """
    Returns an integrated map of some signal along the line of sight. This routine
    assumes that the pressure within a shell has a power law distribution.
    
    Parameters
    __________
    r_bins    : class:`numpy.ndarray`
       The radial bins (in radians)
    vals      : class:`numpy.ndarray`
       Pressure for each bin used
    theta     : class:`numpy.ndarray`
       An array of radii (in radian) in the map, which will be used for gridding the model
    alphas    : array-like
       An array of power laws (indices) for 3d pressure distribution
    shockxi   : float
       Polar tapering, if used in a shock model.
    finite    : bool
        Set this keyword if you do NOT want to integrate to infinity.
    narm      : bool
       Normalize at R_min (within a bin)
    strad     : bool, optional
       STrict RADii. When using a shock model (e.g. Abell 2146), where specific radii, especially inner radii are defined, this keyword should be set! Note that if the finite keyword is set, then this does not need to be set. 
    negvals   : class:`numpy.ndarray`, optional
       None by default. Otherwise, set as boolean array, same length as r_bins

    Returns
    -------
    out       : class:`numpy.ndarray`
       Map convolved with the beam.          
    """
    if finite == False:
        iadj = 0
        if np.min(r_bins) != 0 and strad == False:
            mybins=np.append([0],r_bins)
            if len(mybins) == 3:
                mybins=np.append(mybins,[-1])
            else:
                mybins[-1]=-1
        else:
            mybins=np.append(r_bins,-1)
    else:
        # This almost looks wrong - but it should be right ( 20Jan2022 )
        # mybins retains the information, but this allows the for loop
        # to go over the correct number of indices.
        mybins = np.asarray(r_bins).copy()
        r_bins = r_bins[:-1]
        iadj   = 1 # 22 Jan 2022 ...OMG
            
#    import pdb; pdb.set_trace()
    nthetas = len(theta)
    integrals = np.zeros((len(r_bins),nthetas))
    if fixalpha == False:
        alphas = np.zeros(len(r_bins))

    badind=0
    for idx, myval in enumerate(r_bins):
        rin=mybins[idx]
        rout=mybins[idx+1]
        mypressure=vals[iadj+idx] # Gah, what a stupid way to do this.
        
        if rin == rout:
            badind = idx
            continue        
        
        if fixalpha == False:                  
            if rin == 0:
                lr=np.log10(mybins[idx+2]/mybins[idx+1])
                lp=np.log10(vals[idx+1]/vals[idx])
                alphas[idx]=-lp/lr
            elif rout == -1:
                lr=np.log10(r_bins[idx]/r_bins[idx-1])
                lp=np.log10(vals[iadj+idx]/vals[idx+iadj-1])
#                    lr=np.log10(mybins[idx]/mybins[idx-1])
#                    lp=np.log10(vals[idx-1]/vals[idx-2])
                alphas[idx]=-lp/lr
                mypressure=vals[idx+iadj-1]
            else:
                lr=np.log10(mybins[idx+1]/mybins[idx])
                lp=np.log10(vals[idx+iadj]/vals[idx+iadj-1])
                alphas[idx]=-lp/lr
                #if alphas[idx] < -20:
                #    import pdb;pdb.set_trace()
 
### Beware of 2.0*shockxi!!! (26 July 2017)
        #import pdb;pdb.set_trace()
        #integrals[idx] = shell_pl(mypressure,alphas[idx]+2.0*shockxi,rin,rout,theta,narm=narm) #R had been in here.
        ### 2*shockxi doesn't seem correct (17 Dec 2021)
        
        #if np.any(np.isnan(alphas)):
        #    import pdb;pdb.set_trace()

        integrals[idx] = shell_pl(mypressure,alphas[idx]+shockxi,rin,rout,theta,narm=narm,tmax=tmax) #R had been in here.

    #import pdb;pdb.set_trace()
    if negvals is None or np.sum(negvals) == 0:
        totals = np.sum(integrals,axis=0)  # This should accurately produce Compton y values.
    else:
        pdint  = integrals.copy()
        #import pdb;pdb.set_trace()
        pdint[negvals,:] *= -1
        totals = np.sum(pdint,axis=0)  # This should accurately produce Compton y values.
        #print("Hi")

    if badind > 0:
        alphas    = np.delete(alphas,badind)
        integrals = np.delete(integrals,badind,0)
        
    return totals,alphas,integrals

def iter_grid_profile(integrals, myrs, theta_range, xymap, geoparams=[0,0,0,1,1,1,0,0],axis='z'):
    """
    This largely copies the functionality of grid_profile, but is designed to be much faster for
    iterative applications (same geoparams)

    :param integrals: array of los-integrated values (i.e. SB profiles)
    :type integrals: class:`numpy.ndarray`
    :param myrs: array of bin radii
    :type myrs: array-like
    :param theta_range: profile of radii
    :type theta_range: class:`numpy.ndarray`
    :param xymap: tuple of arrays of x and y coordinates
    :type xymap: tuple
    :param geoparams   :  [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    :type geoparams    : array-like
    :param axis        :  scale about x, y, or z
    :type axis         : str
    """

    ### Get new grid:
    (x,y) = xymap
    x,y = rot_trans_grid(x,y,geoparams[0],geoparams[1],geoparams[2])
    x,y = get_ell_rads(x,y,geoparams[3],geoparams[4])
    radmap = np.sqrt(x**2 + y**2)
    theta = radmap*(u.arcsec).to("radian");  theta_min = np.min(theta_range)
    bi=np.where(theta < theta_min);   theta[bi]=theta_min
    nx, ny = theta.shape

    rsRads = myrs / np.min(myrs)   # This should be unitless from Python's perspective, but really in arcseconds.
    
    ### And a couple more *necessary* modification:
    ### Where we want to scale it by a certain r_bin, given in radians. We also want to scale by "Ella", if axis='x':
    if axis == 'x':
        xell = (x/(geoparams[3]*np.min(myrs)))*(u.arcsec).to("radian") # x is initially presented in arcseconds
        modmap = geoparams[5]*(xell**2)**(geoparams[6]) # Consistent with model creation??? (26 July 2017)
    if axis == 'y':
        yell = (y/(geoparams[4]*np.min(myrs)))*(u.arcsec).to("radian") # x is initially presented in arcseconds
        modmap = geoparams[5]*(yell**2)**(geoparams[6]) # Consistent with model creation??? (26 July 2017)
    if axis == 'z':
        modmap = np.zeros((nx,ny)) + geoparams[5]      # Just the plain old LOS elongation factor

    mymap = np.zeros(x.shape)
    for profile, myscale in zip(integrals, rsRads):
        
        fint = interp1d(theta_range, profile, bounds_error = False, fill_value = 0)
        #map = np.float64(fint(theta.reshape(nx * ny))) # Type 17 = float? (Implicitly float 32?)
        map = fint(theta.reshape(nx * ny)) # Type 17 = float? (Implicitly float 32?)
        map = map.reshape(nx,ny)
        map = map * modmap * (myscale**(-2*geoparams[6]))
        mymap+=map
        
    if geoparams[7] > 0:
        angmap = np.arctan2(y,x)
        bi = np.where(abs(angmap) > geoparams[7]/2.0)
        mymap[bi] = 0.0

    return mymap

def iter_grid_profile_v2(integrals, myrs, theta_range, xymap, geoparams=[0,0,0,1,1,1,0,0],axis='z',
                         xyinas=True):
    """
    This largely copies the functionality of grid_profile, but is designed to be much faster for
    iterative applications (same geoparams)

    :param integrals: array of los-integrated values (i.e. SB profiles)
    :type integrals: class:`numpy.ndarray`
    :param myrs: array of bin radii
    :type myrs: array-like
    :param theta_range: profile of radii
    :type theta_range: class:`numpy.ndarray`
    :param xymap: tuple of arrays of x and y coordinates
    :type xymap: tuple
    :param geoparams   :  [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    :type geoparams    : array-like
    :param axis        :  scale about x, y, or z
    :type axis         : str
    :param xinas       :  xymap is in arcseconds?
    :type xinas        : bool
    """

    a2r    = (u.arcsec).to("radian")
    conv   = a2r*1.0 if xyinas else 1.0
    ### Get new grid:
    (x,y) = xymap
    x,y = rot_trans_grid(x,y,geoparams[0],geoparams[1],geoparams[2])
    x,y = get_ell_rads(x,y,geoparams[3],geoparams[4])
    radmap = np.sqrt(x**2 + y**2)
    theta = radmap*conv;  theta_min = np.min(theta_range)
    bi=np.where(theta < theta_min);   theta[bi]=theta_min
    #nx, ny = theta.shape

    rsRads = myrs / np.min(myrs)   # This should be unitless from Python's perspective, but really in arcseconds.
    ### And a couple more *necessary* modification:
    ### Where we want to scale it by a certain r_bin, given in radians. We also want to scale by "Ella", if axis='x':
    if axis == 'x':
        xell = (x/(geoparams[3]*np.min(myrs)))*conv # x is initially presented in arcseconds
        modmap = geoparams[5]*(xell**2)**(geoparams[6]) # Consistent with model creation??? (26 July 2017)
    if axis == 'y':
        yell = (y/(geoparams[4]*np.min(myrs)))*conv # x is initially presented in arcseconds
        modmap = geoparams[5]*(yell**2)**(geoparams[6]) # Consistent with model creation??? (26 July 2017)
    if axis == 'z':
        modmap = np.zeros(theta.shape)*0.0 + geoparams[5]      # Just the plain old LOS elongation factor

    Int_Prof = np.zeros((integrals.shape[1]))

    for profile, myscale in zip(integrals, rsRads):
        Int_Prof+=  (myscale**(-2*geoparams[6])) * profile

    #import pdb;pdb.set_trace()
        
    fint = interp1d(np.float64(theta_range),np.float64(Int_Prof), bounds_error = False, fill_value = 0)
    #mymap = np.float64(fint(theta.reshape(theta.size))) # Type 17 = float? (Implicitly float 32?)
    mymap = fint(theta.reshape(theta.size)) # Type 17 = float? (Implicitly float 32?)
    mymap = mymap.reshape(theta.shape)
    mymap = mymap * modmap 
        
    if geoparams[7] > 0:
        angmap = np.arctan2(y,x)
        bi = np.where(abs(angmap) > geoparams[7]/2.0)
        mymap[bi] = 0.0

    return mymap

##############################################################
##############################################################
##############################################################


def shell_pl(epsnot,sindex,rmin,rmax,radarr,c=1.0,ff=1e-3,epsatrmin=0,
             narm=False,tmax=0):

    """
    The heart of this code. This routine calculates analytic los integrals depending on the case.

    :param epsnot: The normalization factor. The default behavior is for this to be defined at RMAX, the outer edge of a sphere or shell. If you integrate to infinity, then this should be defined at RMIN. And of course, RMIN=0, and RMAX as infinity provides no scale on which to define EPSNOT. See the optional variable EPSATRMIN.
    :type epsnot: float
    :param sindex: "Spectral Index". That is, the power law (without the minus sign) that the "emissivity" follows within your bin. If you want to integrate to infinity, you must have SINDEX > 1. All other cases can handle any SINDEX value.
    :type sindex: float
    :param rmin: Minimum radius for your bin. Can be 0.
    :type rmin: float
    :param rmax: Maximum radius for your bin. If you wish to set this to infinity, then set it to a negative value.
    :type rmax: float
    :param radarr: A radial array of projected radii (same units as RMIN and RMAX) for which projected values will be calculated. If the innermost value is zero, its value, in the scaled radius array will be set to FF.
    :type radarr: class:`numpy.ndarray`
    :param c: The scaling axis for an ellipse along the line of sight. Default is 1.0
    :type c: float
    :param ff: Fudge Factor, but more of a thresholding factor. Default is 1e-3.
    :type ff: float
    :param epsatrmin: Set this to a value greater than 0 if you want EPSNOT to be defined at RMIN. This automatically happens if RMAX<0
    :type epsatrmint: float
    :param narm: Normalized At R_Min. This option specifies that you have *already* normalized the bins at R_Min (for a shell case). The other two cases are strictly imposed where the normalization is set. The default is False, because that is just how I started using this.
    :type narm: bool
    :param tmax: Maximum theta (from the nose - of, say, a shock).
    :type tmax: float

    NOTE: If RMIN = 0 and RMAX < 0, then this program will return 0.
    """

    ##############################################################
    ### OUTPUTS:
    #
    # PLINT     - PLINT is the integration along the z-axis (line of sight) for
    #             an ellipsoid (a sphere) where the "emissivity" is governed by
    #             a power law. The units are thus given as the units on EPSNOT
    #             times the units on RADARR (and therefore RMIN and RMAX).
    #
    #             It is then dependent on you to make the appropriate
    #             conversions to the units you would like.
    # 
    ##############################################################
    ### Perform some double-checks.

    if rmin < 0:
        print('found rmin < 0; setting rmin equal to 0')
        rmin = 0

    rrmm = (radarr==np.amin(radarr))
    if (radarr[rrmm] == 0) and (sindex > 0):
        radarr[rrmm]=ff

    ##############################################################
    ### Determine the appropriate case (and an extra double check)

    if rmax < 0:
        if rmin == 0:
            scase=3
        else:
            scase=2
            epsatrmin=1
    else:
        if rmin == 0:
            scase=0
        else:
            if rmin < rmax:
                scase=1
                epsatrmin=1
            else:
                print('You made a mistake: rmin > rmax; sending to infty integration.')
                ### If a mistake is possible, it will happen, eventually.
                scase=3

    ### Direct program to appropriate case:
    shellcase = {0: plsphere, # You are integrating from r=0 to R (finite)
                 1: plshell,  # You are integrating from r=R_1 to R_2 (finite)
                 2: plsphole, # You are integrating from r=R (finite, >0) to infinity
                 3: plinfty,  # You are integrating from r=0 to infinity
                 }

    ##############################################################
    ### Redo some numbers to agree with hand-written calculations

    p = sindex/2.0 # e(r) = e_0 * (r^2)^(-p) for this notation / program

    ### In a way, I actually like having EPSNORM default to being defined at RMIN
    ### (Easier to compare to hand-written calculations.

    if scase ==1 and narm == False:
        epsnorm=epsnot*(rmax/rmin)**(sindex)
    else:
        epsnorm=epsnot

    ### Prefactors change a bit depending on integration method.
    ### These are the only "pre"factors common to all (both) methods.
    prefactors=epsnorm*c
    ### Now integrate for the appropriate case
    myintegration = shellcase[scase](p,rmin,rmax,radarr,tmax=tmax)
  
    answer = myintegration*prefactors  ## And get your answer!
  
    return answer

##############################################################
##### Integration cases, as directed above.              #####
##############################################################

def plsphere(p,rmin,rmax,radarr,tmax=0):
    """
    Analytic los integral for a full sphere.

    :param p: "Spectral Index". That is, the power law (without the minus sign) that the "emissivity" follows within your bin. If you want to integrate to infinity, you must have SINDEX > 1. All other cases can handle any SINDEX value.
    :type p: float
    :param rmin: Minimum radius for your bin. This is 0 for the sphere.
    :type rmin: float
    :param rmax: Maximum radius for your bin. If you wish to set this to infinity, then set it to a negative value.
    :type rmax: float
    :param radarr: A radial array of projected radii (same units as RMIN and RMAX) for which projected values will be calculated. If the innermost value is zero, its value, in the scaled radius array will be set to FF.
    :type radarr: class:`numpy.ndarray`
    :param tmax: Maximum theta (from the nose - of, say, a shock).
    :type tmax: float
    """

    
    c1 = radarr<=rmax              # condition 1
    c2 = radarr>rmax               # condition 2
#    c1 = np.where(radarr<=rmax)     # condition 1
#    c2 = np.where(radarr>rmax)      # condition 2
    sir=(radarr[c1]/rmax)           # scaled radii
    isni=((2.0*p==np.floor(2.0*p)) and (p<=1)) # Special cases -> "method 2"
    plinn = sir**(1.0-2.0*p)
    if tmax > 0:
        ct = np.cos(tmax)
        igi = (sir < ct)
        sir[igi] = ct
    if isni:
      tmax=np.arctan(np.sqrt(1.0 - sir**2)/sir)   # Theta max
      plint=myredcosine(tmax,2.0*p-2.0)*plinn*2.0 # Integration + prefactors
    else:
      cbf=(sps.gamma(p-0.5)*np.sqrt(np.pi))/sps.gamma(p) # complete beta function
      ibir=myrincbeta(sir**2,p-0.5,0.5)               # incomplete beta function
      plint=plinn*(1.0-ibir)*cbf     # Apply appropriate "pre"-factors

    myres=radarr*0          # Just make my array (unecessary?)
    myres[c1]=plint         # Define values for R < RMIN

    #if np.any(np.isnan(myres)):
    #    import pdb;pdb.set_trace()
    
    return myres*rmax               # The results we want

def plshell(p,rmin,rmax,radarr,tmax=0):
    """
    Analytic los integral for a spherical shell.

    :param p: "Spectral Index". That is, the power law (without the minus sign) that the "emissivity" follows within your bin. If you want to integrate to infinity, you must have SINDEX > 1. All other cases can handle any SINDEX value.
    :type p: float
    :param rmin: Minimum radius for your bin. This is 0 for the sphere.
    :type rmin: float
    :param rmax: Maximum radius for your bin. If you wish to set this to infinity, then set it to a negative value.
    :type rmax: float
    :param radarr: A radial array of projected radii (same units as RMIN and RMAX) for which projected values will be calculated. If the innermost value is zero, its value, in the scaled radius array will be set to FF.
    :type radarr: class:`numpy.ndarray`
    :param tmax: Maximum theta (from the nose - of, say, a shock).
    :type tmax: float
    """    
    c1 = radarr<=rmax              # condition 1
    c2 = radarr[c1]<rmin           # condition 2
    c3 = radarr<rmin               # c1[c2] as I would expect in IDL
#    c1 = np.where(radarr<=rmax)     # condition 1
#    c2 = np.where(radarr[c1]<rmin)  # condition 2
    sir=(radarr[c1]/rmin)           # scaled inner radii
    sor=(radarr[c1]/rmax)           # scaled outer radii

    plinn=sir**(1.0-2.0*p)                 # Power law term for inner radii
    
    if tmax > 0:
        ct = np.cos(tmax)
        igi = (sir < ct)
        ogi = (sor < ct)
        sir[igi] = ct
        sor[ogi] = ct
    isni=((2.0*p==np.floor(2.0*p)) and (p<=1)) # Special cases -> "method 2"
    myres=radarr*0                  # Just make my array (unecessary?)
    if isni:
      tmxo=np.arctan(np.sqrt(1.0 - sor**2)/sor)         # Theta max...outer circle
      tmxi=np.arctan(np.sqrt(1.0 - sir[c2]**2)/sir[c2]) # Theta max...inner circle
      plint=myredcosine(tmxo,2.0*p-2.0)              # Integrate for outer circle.
      plint[c2]-=myredcosine(tmxi,2.0*p-2.0) # Integrate and subtract inner circle
#      myres[c1]=plint*(sor**(1.0-2.0*p))*2.0    # Pre-(24 July 2017) line.
      myres[c1]=plint*plinn*2.0    # Apply appropriate "pre"-factors
      
    else:
      cbf=(sps.gamma(p-0.5)*np.sqrt(np.pi))/sps.gamma(p) # complete beta function
      ibir=myrincbeta(sir[c2]**2,p-0.5,0.5) # Inc. Beta for inn. rad.
      ibor=myrincbeta(sor**2,p-0.5,0.5)     # Inc. Beta for out. rad.
      #plinn=(sir**(1.0-2.0*p))                 # Power law term for inner radii
      myres[c1]=plinn*(1.0-ibor)*cbf           # Define values for the enclosed circle
#      import pdb;pdb.set_trace()
#      myres[c1[c2]]=plinn[c2]*(ibir-ibor[c2])*cbf # Correct the values for the
### Changed this March 9, 2018:
      myres[c3]=plinn[c2]*(ibir-ibor[c2])*cbf # Correct the values for the 
      # inner circle
      
    #if np.any(np.isnan(myres)):
    #    import pdb;pdb.set_trace()
                                               
    return myres*rmin                          # The results we want

def plsphole(p,rmin,rmax,radarr,tmax=0):
    """
    Analytic los integral for out to infinity, but missing a spherical core.

    :param p: "Spectral Index". That is, the power law (without the minus sign) that the "emissivity" follows within your bin. If you want to integrate to infinity, you must have SINDEX > 1. All other cases can handle any SINDEX value.
    :type p: float
    :param rmin: Minimum radius for your bin. This is 0 for the sphere.
    :type rmin: float
    :param rmax: Maximum radius for your bin. If you wish to set this to infinity, then set it to a negative value.
    :type rmax: float
    :param radarr: A radial array of projected radii (same units as RMIN and RMAX) for which projected values will be calculated. If the innermost value is zero, its value, in the scaled radius array will be set to FF.
    :type radarr: class:`numpy.ndarray`
    :param tmax: Maximum theta (from the nose - of, say, a shock).
    :type tmax: float
    """    
    
    if p <= 0.5:
        return radarr*0 - 1.0e10

    else:
        c1 = radarr<rmin               # condition 1
        c2 = radarr>=rmin              # condition 2
        #      c1 = np.where(radarr<rmin)     # condition 1
        #      c2 = np.where(radarr>=rmin)    # condition 2
        sr=(radarr/rmin)               # scaled radii
        plinn=(sr**(1.0-2.0*p))          # Power law term
        if tmax > 0:
            ct = np.cos(tmax)
            gi = (sr < ct)
            sr[gi] = ct
        cbf=(sps.gamma(p-0.5)*np.sqrt(np.pi))/sps.gamma(p) # complete beta function
        ibor=myrincbeta(sr[c1]**2,p-0.5,0.5) # Inc. Beta for out. rad.
        myres=radarr*0                 # Just make my array (unecessary?)
        myres[c1]=plinn[c1]*ibor*cbf     # Define values for R < RMIN
        myres[c2]=plinn[c2]*cbf          # Define values for R > RMIN
        #if np.any(np.isnan(myres)):
        #    import pdb;pdb.set_trace()
        
        return myres*rmin

def plinfty(p,rmin,rmax,radarr,tmax=None):
    """
    Analytic los integral for infinity. It doesn't work. Returns 0.

    :param p: "Spectral Index". That is, the power law (without the minus sign) that the "emissivity" follows within your bin. If you want to integrate to infinity, you must have SINDEX > 1. All other cases can handle any SINDEX value.
    :type p: float
    :param rmin: Minimum radius for your bin. This is 0 for the sphere.
    :type rmin: float
    :param rmax: Maximum radius for your bin. If you wish to set this to infinity, then set it to a negative value.
    :type rmax: float
    :param radarr: A radial array of projected radii (same units as RMIN and RMAX) for which projected values will be calculated. If the innermost value is zero, its value, in the scaled radius array will be set to FF.
    :type radarr: class:`numpy.ndarray`
    :param tmax: Maximum theta (from the nose - of, say, a shock).
    :type tmax: float
    """    
    sr=(radarr)                      # scaled radii
    cbf=(sps.gamma(p-0.5)*np.sqrt(np.pi))/sps.gamma(p) # complete beta function
    plt=(sr**(1.0-2.0*p))          # Power law term

### There is no scaling to be done: RMIN=0; RMAX=infinity...
### This is madness, but if you can set >>SOME<< scaling radius, this can work.
### However, the practical implementation of this is not foreseen / understood
### how it should look. Therefore, for now, I will return 0.

    return 0       # Scale invariant. Right. Fail.


def myrincbeta(x,a,b):
    """
    compute the regularized incomplete beta function.
    .. math::

        B(x;a,b) \\equiv \\int_0^x u^{a-1} (1-u)^{b-1} du

    :param x: Reflective of the radius out to which this is computed
    :type x: float
    :param a: Nameless parameter?
    :type a: float
    :param b: Nameless parameter?
    :type b: float
    """
    if a < 0:
        cbf=(sps.gamma(a)*sps.gamma(b))/sps.gamma(a+b)
        res = (x**a * (1.0-x)**b) / (a * cbf)
        #if np.any(np.isnan(res)):
        #    import pdb;pdb.set_trace()
        return myrincbeta(x,a+1.0,b) + res
    else:
        #      cbf=(sps.gamma(a)*sps.gamma(b))/sps.gamma(a+b)
        cbf=1.0 # sps.betainc is the regularized inc. beta fun.
        res=(sps.betainc(a,b,x) / cbf)
        return res
    
def myredcosine(tmax,n):
    """
    computes 
    .. math::

    \int_0^tmax cos^n(x) dx

    :param tmax: :math:`\\theta_{max}`
    :type tmax: float
    :param n: exponent
    :type n: int

    """

    if n < -2:
        res=np.cos(tmax)**(n+1)*np.sin(tmax)/(n+1) 
        return myredcosine(tmax,n+2)*(n+2)/(n+1) - res
    else:
        if n == 0:
            res=tmax
        if n == -1:
            res=np.log(np.absolute(1.0/np.cos(tmax) + np.tan(tmax)) )
        if n == -2:
            res=np.tan(tmax) 

    return res

def ycyl_prep(Int_Pres,theta_range):
    """
    Just preparing some numbers.

    :param Int_Pres: integrated pressures
    :type Int_Pres: class:`numpy.ndarray`
    :param theta_range: radii
    :type theta_range: class:`numpy.ndarray`

    """

    lnp = np.log(Int_Pres)
    ltr = np.log(theta_range)

    alpha = (np.roll(lnp,-1) - lnp ) / (np.roll(ltr,-1) - ltr )
    k     = Int_Pres / theta_range**alpha

    return alpha,k

def handle_warning(message, category, filename, lineno, file=None, line=None):
    print('A warning occurred:')
    print(message)
    print('Do you wish to continue?')

    while True:
        response = input('y/n: ').lower()
        if response not in {'y', 'n'}:
            print('Not understood.')
        else:
            break

    if response == 'n':
        import pdb;pdb.set_trace()
        #raise category(message)



import numpy as np                         # A useful package...
import emcee, os, corner                   # A list of modules to import as-is
from astropy.io import fits                # To read/write fits
import M2_ProposalTools.analytic_integrations as ai         # Integrates ellipsoidal power law distributions
import astropy.units as u                  # Allows variables (quantities) to have units
import M2_ProposalTools.WorkHorse as WH    #
from scipy.optimize import curve_fit
from importlib import reload
from scipy.interpolate import interp1d     
from astropy.wcs import WCS
import matplotlib.pyplot as plt
import scipy.special as scs
import M2_ProposalTools.FilterImages as FI

WH=reload(WH)
    

def fit_spherical_model(z,M500,hdul,model="NP",pink=True,alpha=2,knee=1.0/120.0,nkbin=100,fwhm=9.0,nsteps=1000,
                        y2k=-3.4,uKinput=True,ySph=True,YMrel="A10",outdir="/tmp/",outbase="NP_fit_corner.png",
                        size=3.5,Dist=False,plotPin=True,adPhys=True,WIKID=False,fit_cen=True,addNoise=True):

    """
    :param z: Redshift
    :type z: float
    :param M500: :math:`M\_{500}`
    :type M500: quantity (mass units)
    :param hdul: Astropy/fits HDU list (Make sure wtmap and img have appropriate units)
    :type hdul: list of HDU class objects 
    :param model: Type of model. Can be "NP" (for non-parametric), "GNFW" for generalized NFW, or "BETA" for a beta model.
    :type model: str
    :param pink: noise realization is pink noise
    :type pink: bool
    :param alpha: power-law index for red-noise part of pink noise. Default is 2.
    :type alpha: float
    :param knee: where, in inverse arcseconds, does the knee occur.
    :type knee: float
    :param nkbin: number of k-bins used to generate spectrum for pink noise.
    :type nkbin: int
    :param fwhm: Smoothing kernel for filtered map.
    :type fwhm: float
    :param nsteps: Number of steps to use in MCMC.
    :type nsteps: int
    :param y2k: Conversion factor between Compton y and Kelvin_RJ (for MUSTANG-2). Default is -3.4.
    :type y2k: float
    :param uKinput: The input map is taken to be in units of microKelvin_RJ if set.
    :type uKinput: bool
    :param ySph: Calculate :math:`y_{Sph}` if set; otherwise calculate :math:`y_{cyl}`. Default is True.
    :type ySph: bool
    :param YMrel: Which Y-M relation to use? Default is "A10".
    :type YMrel: str
    :param outdir: A string indicating the output directory, including trailing "/".
    :type outdir: str
    :param outbase: filename, without the directory path.
    :type outbase: str
    :param size: size of Lissajous daisy scan used (for transfer function). Options are 2.5, 3.0, 3.5, 4.0, 4.5, or 5.0
    :type size: float.
    :param fit_cen: Fit for a center? Default is True
    :type fit_cen: bool
    
    """

    ### First, we need to establish the radius out to which we fit and how many bins we want.
    ### This relies on how good our data are.
    
    conv         = 1e-6/y2k if uKinput else 1.0/y2k 
    hdul[0].data = hdul[0].data*conv
    hdul[1].data = hdul[1].data / ( conv**2 )
    hdu4snr      = hdul.copy()
    sf           = WH.get_smoothing_factor(fwhm=fwhm)
    if addNoise:
        print("Making SNR map")
    else:
        pixsize      = WH.get_pixarcsec(hdul)
        SigImg       = FI.fourier_filtering_2d(hdul[0].data,"gauss",fwhm/pixsize)
        SigWts       = FI.fourier_filtering_2d(hdul[1].data,"gauss",fwhm/pixsize)
        SigWts      *= (sf/pixsize)**2
        print(sf/pixsize,conv)
        hdu4snr[0].data = SigImg
        hdu4snr[1].data = SigWts
        sf           = 1.0
    SNRmap       = WH.get_SNR_map(hdu4snr)
    SNRhdu       = fits.PrimaryHDU(SNRmap,header=hdul[0].header)
    SNRhdu.writeto(outdir+"SNRmap.fits",overwrite=True)
    CosmoPars    = WH.get_cosmo_pars(z)
    pixsize      = WH.get_pixarcsec(hdul)
    intSNR,xc,yc,xymap = get_int_SNR(SNRmap,pixsize,maxRad=2.0,bv=120.0,SNRthresh=1.0)
    if addNoise == False:
        intSNR = np.sqrt(intSNR*2)
    print("Integrated SNR taken to be: ",intSNR)
    print("Adopting a center of ",xc,yc)

    if addNoise:
        Noise        = WH.get_noise_realization(hdul,pink=pink,alpha=alpha,knee=knee,nkbin=nkbin,fwhm=fwhm)
    else:
        Noise        = np.zeros(hdul[0].data.shape)
    # Modify HDU list to have mock-data and correctly weighted pixels.
    # Want to work in Compton y for fitting.

    ###############################################################################################

    efv        = get_emcee_fit_vars(CosmoPars,M500,intSNR,xc,yc,pixsize,outdir,fit_cen=fit_cen,
                                    MinRes=pixsize/2.0,model=model,ySph=ySph,YMrel=YMrel,size=size,
                                    Dist=Dist,WIKID=WIKID)
    mask         = automated_mask(xymap,CosmoPars,efv)

    hdul[0].data = hdul[0].data + Noise
    hdul[1].data = hdul[1].data*mask / ( sf**2 )

    hdul.writeto(outdir+"InputHDU.fits",overwrite=True)
    
    run_emcee(hdul,CosmoPars,efv,xymap,outdir+outbase,BSerr=False,nsteps=nsteps,plotPin=plotPin,adPhys=adPhys)

def automated_mask(xymap,cosmo_pars,efv):

    rmap         = WH.make_rmap(xymap)
    r500         = efv["Theta500"]*3600*180/np.pi # arcseconds

    goodind      = (rmap < 1.5*r500)
    mask         = rmap*0
    mask[goodind]= 1.0

    return mask

def get_emcee_fit_vars(cosmo_pars,M500,SNRint,xc,yc,pixsize,outdir,
                       MinRes=1.0,model="NP",ySph=True,YMrel="A10",nb_theta_range=150,SNRperbin=5.0,
                       n_at_rmin=False,fit_mnlvl=True,fit_cen=True,fit_geo=False,size=3.5,Dist=False,
                       WIKID=False):
    """
    :param cosmo_pars: a dictionary of cosmological parameters
    :type cosmo_pars: dict
    :param M500: :math:`M\\_{500}`
    :type M500: quantity (mass units)
    :param SNRint: Integrated SNR (how many sigma detection).
    :type SNRint: float
    :param xc: x-centroid, in pixels
    :type xc: float
    :param yc: y-centroid, in pixels
    :type yc: float
    :param pixsize: pixel size, in arcseconds
    :type pixsize: float
    :param outdir: A string indicating the output directory, including trailing "/".
    :type outdir: str
    :param MinRes: minimum resolution (in defining radial profile). Default is 1 arcsecond.
    :type MinRes: float
    :param model: Type of model. Can be "NP" (for non-parametric), "GNFW" for generalized NFW, or "BETA" for a beta model.
    :type model: str
    :param ySph: Calculate :math:`y_{Sph}` if set; otherwise calculate :math:`y_{cyl}`. Default is True.
    :type ySph: bool
    :param YMrel: Which Y-M relation to use? Default is "A10".
    :type YMrel: str
    :param nb_theta_range: number of points in theta_range array.
    :type nb_theta_range: int
    :param SNRperbin: minimum desired SNR per bin. Default is 4.0.
    :type SNRperbin: float
    :param n_at_rmin: Normalize at r_min. Default is False. (Unless you know what you are doing, leave this.)
    :type n_at_rmin: bool
    :param fit_mnlvl: Fit for a mean level? Default is True
    :type fit_mnlvl: bool
    :param fit_cen: Fit for a center? Default is True
    :type fit_cen: bool
    :param fit_geo: Fit for a an elliptical geometry? Default is False
    :type fit_geo: bool
    :param size: size of Lissajous daisy scan used (for transfer function). Options are 2.5, 3.0, 3.5, 4.0, 4.5, or 5.0
    :type size: float.
    
    """
    Theta500    = WH.Theta500_from_M500_z(M500,cosmo_pars["z"])
    minpixrad   = (MinRes*u.arcsec).to('rad')
    
    tnx         = [minpixrad.value,10.0*Theta500]  # In radians
    thetas      = np.logspace(np.log10(tnx[0]),np.log10(tnx[1]), nb_theta_range)
    
    Pdl2y       = WH.get_Pdl2y(cosmo_pars["z"],cosmo_pars["d_ang"])
    
    m500s       = WH.r2m_delta(thetas*cosmo_pars["d_a"],cosmo_pars["z"],delta=500)
    y500s       = WH.y_delta_from_mdelta(m500s,cosmo_pars["z"],delta=500,ycyl=(not ySph),YMrel=YMrel,h70=cosmo_pars["h70"])
    #geom     = [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    default_geo = [0,0,0,1,1,1,0,0]

    nBins       = int(np.round( (SNRint/SNRperbin)**2 ))
    if nBins < 4:
        print(SNRint,SNRperbin)
        print("Oh no, nBins is less than 4! Your data is probably not sufficient for this analysis")
        nBins   = 4
        
    radmin      = 5.0 *np.pi / (180*3600)                # 5" in radians
    radminmax   = np.array( [radmin,Theta500] )          # Between 5" and R500, in radians
    nMax        = np.floor(Theta500*3600*180/np.pi/12.0) # I want a bit larger than 10" bins.
    if nBins >= nMax:
        radminmax   = np.array( [radmin,Theta500*1.5] )          # Between 5" and R500, in radians
        nBins       = int(np.round(nMax))-2
    print(radminmax,nBins,SNRint,SNRperbin)
    #import pdb;pdb.set_trace()
    
    bins        = improved_bin_spacing(radminmax,nBins,mway=False) # in radians
    binkpc      = bins*cosmo_pars["d_a"]
    
    myval     = []   # 
    sbepos    = []   # Boolean array indicating whether myval should be positive or not.
    labels    = []
    uppbnd    = []
    lowbnd    = []
    
    ######################################################################################
    if fit_mnlvl:
        myval.append(1e-6)
        sbepos.append(False)
        labels.append("MnLvl")
        uppbnd.append(1e-5)
        lowbnd.append(-1e-5)

    if model == "NP":
        PresProf  = WH.a10_from_m500_z(M500, cosmo_pars["z"],binkpc*u.kpc,Dist=False)
        uless_p = (PresProf*Pdl2y).decompose().value
        sbepos.extend([True]*nBins)
        #NPlabels = [rf'''P_{sub}''' for sub in range(nBins)]
        NPlabels = [rf'''P$_{{{sub}}}$''' for sub in range(nBins)]
        labels.extend(NPlabels)
    if model == 'BETA':
        r500, p500 = WH.R500_P500_from_M500_z(M500, cosmo_pars["z"])
        uless_p = np.array([p500.to('keV / cm**3').value*10.0 ,Theta500/10.0,1.0])
        sbepos.extend(np.ones((len(uless_p)),dtype=bool))
        BetaLabels = [r"P$_0$",r"r$_c$",r"$\beta$"]
        labels.extend(BetaLabels)
    myval.extend(uless_p)
    pmult = np.ones(uless_p.shape)*10
    pmult[-1] *= 2
    uppbnd.extend(uless_p*pmult)
    lowbnd.extend(uless_p/pmult)

    if fit_cen:
        fcval = np.array([1.0,1.0])
        myval.extend(fcval)
        sbepos.extend([False,False])
        centLabels = [r"$\delta_x$",r"$\delta_y$"]
        labels.extend(centLabels)
        uppbnd.extend(fcval*20)
        lowbnd.extend(-fcval*20)

    ######################################################################################

    zstr   = "{:.1f}".format(cosmo_pars["z"]).replace(".","z")
    Mstr   = "{:.1f}".format(M500.to("M_sun").value/1e14).replace(".","m")


    efv         = {"Thetas":thetas,"y500s":y500s,"model":model,"ySph":ySph,"Pdl2y":Pdl2y,"geo":default_geo,
                   "M500":M500,"Theta500":Theta500,"YMrel":YMrel,"bins":bins,"nBins":nBins,"fit_mnlvl":fit_mnlvl,
                   "fit_cen":fit_cen,"fit_geo":fit_geo,"narm":n_at_rmin,"fixalpha":False,"finite":False,
                   "pinit":myval,"sbepos":sbepos,"pixsize":pixsize,"labels":labels,"outdir":outdir,
                   "size":size,"Mstr":Mstr,"zstr":zstr,"lowbnd":lowbnd,"uppbnd":uppbnd,"Dist":Dist,
                   "WIKID":WIKID}    

    return efv

def run_emcee(hdul,cosmo_pars,efv,xymap,outfile,BSerr=False,nsteps=1000,plotPin=True,adPhys=True):
    """
    Run the actual fitting procedure.

    :param hdul: an input fits HDUList with extension=0 being the data and extension=1 being the weights.
    :type hdul: list
    :param cosmo_pars: Dictionary of cosmological parameters
    :type cosmo_pars: dict
    :param efv: Dictionary of emcee fitting variables
    :type efv: dict
    :param xymap: tuple of arrays of x- and y-coordinates.
    :type xymap: tuple
    :param outfile: output file (full path)
    :type outfile: str
    :param BSerr: Bootstrap error?
    :type BSerrL bool
    :param nsteps: Number of fitting steps. Default is 1000.
    :type nsteps: int

    """

    # Call this once outside of the MCMC
    tab = WH.get_xfertab(efv["size"],WIKID=efv["WIKID"])

    def lnlike(pos):                          ### emcee_fitting_vars
        outmap,yint,outalpha,mnlvl = make_skymodel_map(hdul,pos,cosmo_pars,efv,xymap)
        #model   = WH.lightweight_filter_ptg(outmap,efv["size"],efv["pixsize"],WIKID=efv["WIKID"]) + mnlvl
        # Then just apply the transfer function (rather than reading in the file each iteration)
        model   =  FI.apply_xfer(outmap,tab,efv["pixsize"]) + mnlvl
        loglike = 0.0
        data    = hdul[0].data
        weights = hdul[1].data       # Best to mask weights (not use the entire map!)
        loglike-= 0.5 * (np.sum(((model - data)**2) * weights))

        return loglike, outalpha,yint

    def fprior(pos):

        prespos = pos[efv["sbepos"]]
        if all([param > 0.0 for param in prespos]):
            return True
        else:
            return False

    def lnprior(pos,outalphas,ycyl):        

        plike = 0.0
        
        for myouts in outalphas:
            if len(myouts) == 0:
                slopeok = True
            else:
                if myouts[-1] > 2.0: slopeok = True
                if myouts[-1] <= 2.0:
                    slopeok = False
                    break
            
        #prespos = pos[efv.sbepos]
        winbnd = np.less(pos,efv["uppbnd"])*np.greater(pos,efv["lowbnd"])
        
        if all(winbnd) and (slopeok == True):
            return plike
        else:
            return -np.inf
        
        
    def lnprob(pos):
        if fprior(pos):
            likel,outalphas,ycyl = lnlike(pos)
            plike                = lnprior(pos,outalphas,ycyl)
            if np.any(np.isnan(likel)):
                print(pos)
                likel = -np.inf
        else:
            likel = -np.inf
            plike = 0
            ycyl  = [0]
        return likel+plike , [ycyl]

    ndim     = len(efv["pinit"])
    nwalkers = 5*ndim
    p0       = efv["pinit"] * (1 + 3e-2*np.random.randn(nwalkers, ndim))
    efv["nburn"]    = nsteps//20
    efv["nwalkers"] = nwalkers
    efv["ndim"]     = ndim
    efv["p0"]       = p0
    efv["nsteps"]   = nsteps

    sampler  = emcee.EnsembleSampler(nwalkers, ndim,lnprob,threads = 1)
    state    = sampler.run_mcmc(p0, nsteps,progress=True)
    post_mcmc_products(hdul,sampler,cosmo_pars,efv,xymap,outfile,plotPin=plotPin,adPhys=adPhys)
    
 
def post_mcmc_products(hdul,sampler,cosmo_pars,efv,xymap,outfile,plotPin=True,adPhys=True):
    """
    After the fitting, make plots and stuff

    :param hdul: an input fits HDUList with extension=0 being the data and extension=1 being the weights.
    :type hdul: list
    :param sampler: sampler from emcee
    :type sampler: class
    :param cosmo_pars: Dictionary of cosmological parameters
    :type cosmo_pars: dict
    :param efv: Dictionary of emcee fitting variables
    :type efv: dict
    :param xymap: tuple of arrays of x- and y-coordinates.
    :type xymap: tuple
    :param outfile: output file (full path)
    :type outfile: str
    """

    flat_samples = sampler.get_chain(discard=efv["nburn"], thin=efv["nwalkers"],flat=True)
    mysolns = np.array(list(map(lambda v: (v[1], v[2]-v[1], v[1]-v[0]),
                             zip(*np.percentile(flat_samples, [16, 50, 84],
                                                axis=0)))))
    blobarr     = sampler.get_blobs(flat=True,discard=efv["nburn"], thin=efv["nwalkers"])
    blobarr     = blobarr[np.isfinite(blobarr)]        # No need for NANs or infs.
    Yints       = np.percentile(blobarr, [16, 50, 84],axis=0)
    yinteg      = Yints[1]       # This is the median...i.e. the best-fit value
    #print(yinteg)
    #import pdb;pdb.set_trace()
    r500,m500,p500, msys     = WH.rMP500_from_y500(yinteg,cosmo_pars)
    r500l,m500l,p500l, msysl = WH.rMP500_from_y500(Yints[0],cosmo_pars)
    r500u,m500u,p500u, msysu = WH.rMP500_from_y500(Yints[2],cosmo_pars)
    m_unc_array  = np.array([m500l.to("M_sun").value,m500u.to("M_sun").value])
    m_unc_values = (m_unc_array-m500.to("M_sun").value)/1e14
    m_unc        = np.mean(np.abs(m_unc_values))
    m500v        = m500.to("M_sun").value/1e14
    print("###########################################################################")
    print("R500 found to be: ",r500," radians")
    print("M500 found to be: ",m500v,r" \pm ",m_unc," 1e14 M_sun")
    print("P500 found to be: ",p500.to("keV cm**-3"))
    print("Systematic error on M500: ",msys.value," 1e14 M_sun")
    print("###########################################################################")
    #solnstack[:,:,i] = mysolns
    
    print(mysolns)
    npysave = outfile.replace("corner","solutions")
    npysave = npysave.replace(".png",".npy")
    np.save(npysave,mysolns)
    chains_save = npysave.replace("solutions","chains")
    np.save(chains_save,flat_samples)
    if not outfile is None:
        fig = corner.corner(flat_samples, labels=efv["labels"], quantiles=[0.16,0.84])
        fig.savefig(outfile)

    #return mysolns  

    pos     = mysolns[:,0]
    outmap,yint,outalpha,mnlvl = make_skymodel_map(hdul,pos,cosmo_pars,efv,xymap)
    model   = WH.lightweight_filter_ptg(outmap,efv["size"],efv["pixsize"],WIKID=efv["WIKID"]) + mnlvl

    OutHDU1 = fits.PrimaryHDU(outmap,header=hdul[0].header)
    OutHDU2 = fits.ImageHDU(model,header=hdul[1].header)
    OutHDU  = fits.HDUList([OutHDU1,OutHDU2])
    OutName = "FittedModel_"+efv["Mstr"]+"_"+efv["zstr"]+"_"+repr(efv["nsteps"])+"steps.fits"
    OutHDU.writeto(efv["outdir"]+OutName,overwrite=True)

    m500_res = np.abs(np.array([m500v,m_unc_values[0],m_unc_values[1]])) * 1e14
    if np.any(np.isnan(m500_res)) or np.any(np.isinf(m500_res)):
         m500_res = np.ones(3)*1e10
    plot_pressure_profiles(mysolns,efv,cosmo_pars,m500_res,plotPin=plotPin,adPhys=adPhys)

def plot_pressure_profiles(solns,efv,cosmo_pars,m500,myfs=10,plotPin=True,adPhys=True):

    """
    For now, assuming model == "NP"... need to add others

    :param solns: Array of solutions (with uncertainties)
    :type solns: np.ndarray
    :param efv: emcee fitting variables.
    :type efv: dict
    :param m500: Inferred M500 with uncertainties (3-element array)
    :type m500: array-like
    :param myfs: fontsize; default is 10
    :type myfs: float
    """
    
    Pressures = (solns[1:efv["nBins"]+1]/efv["Pdl2y"]).to("keV cm**-3").value         # keV/cm**3
    BinArcmin = efv["bins"]*60*180/np.pi                            # Arcminutes
    Radii     = efv["Thetas"]*60*180/np.pi                          # Arcminutes

    Pminmax   = [np.min(Pressures[:,0])/10.0,np.max(Pressures[:,0])*5]
    #print(Pressures.shape)
    #import pdb;pdb.set_trace()
    PinterFxn = interp1d(np.log(BinArcmin),np.log(Pressures[:,0]),fill_value="extrapolate")
    Pinterp   = np.exp(PinterFxn(np.log(Radii)))

    myfig     = plt.figure(3,figsize=(5,4),dpi=200)
    myfig.clf()
    myax      = myfig.add_subplot(111)
    yerrs     = np.asarray([Pressures[:,2],Pressures[:,1]])
    myax.errorbar(BinArcmin,Pressures[:,0],yerr=yerrs,linestyle="none",color="C0")
    gi1       = (Radii < np.min(BinArcmin))
    myax.plot(Radii[gi1],Pinterp[gi1],"--",color="C0")
    gi2       = (Radii >= np.min(BinArcmin))*(Radii <= np.max(BinArcmin))
    myax.plot(Radii[gi2],Pinterp[gi2],"-",color="C0",label="Recovered Pressure Profile")
    gi3       = (Radii >= np.max(BinArcmin))
    myax.plot(Radii[gi3],Pinterp[gi3],"--",color="C0")
    myax.set_xscale("log")
    myax.set_yscale("log")
    myax.set_ylim(Pminmax)
    myax.set_xlabel(r"Radius (arcmin)",fontsize=myfs)
    myax.set_ylabel(r"P$_e$ (keV cm$^{-3}$)",fontsize=myfs)

    M500_out    = m500[0]*u.Msun
    Theta500    = WH.Theta500_from_M500_z(M500_out,cosmo_pars["z"])
    Arcmin500   = Theta500 * 60 * 180 / np.pi

    
    mYM_pm   = pos_neg_formatter(m500[0],m500[1],m500[2])
    
    xpos = 10**(np.dot(np.log10(myax.get_xlim()),[0.90,0.10]))
    ypos = 10**(np.dot(np.log10(myax.get_ylim()),[0.60,0.40]))
    #print(xpos,ypos)
    myax.axvline(Arcmin500,linestyle="--",color="c")
    myax.text(Arcmin500*0.98,ypos,r"R$_{500}$",rotation=90,color="c",fontsize=myfs)    
    myax.text(xpos,ypos,r'M$_{500}$: '+mYM_pm,fontsize=myfs)

    if plotPin:
        radkpc    = Radii * 60 * cosmo_pars["scale"] * u.kpc
        PresProf  = WH.a10_from_m500_z(efv["M500"], cosmo_pars["z"],radkpc,Dist=efv["Dist"])
        Pin       = PresProf.to("keV cm**-3").value
        myax.plot(Radii,Pin,"--",label="Input Pressure Profile",color="g")

    if adPhys:
        xlims = myax.get_xlim()
        myax.set_xlim(xlims)
        phax = myax.twiny()
        phax.set_xlim([xlim*cosmo_pars["scale"]*60 for xlim in xlims])
        phax.set_xlabel("Radius (kpc)")
        phax.set_xscale("log")
    else:
        myax.set_title("Recovered Pressure Profile (Simulated)",fontsize=myfs)  

    myax.grid()
    myax.legend(fontsize=myfs)
    myfig.tight_layout()
    myfig.savefig(efv["outdir"]+"RecoveredPressureProfile_"+efv["Mstr"]+"_"+efv["zstr"]+".png")
    
def make_skymodel_map(hdul,pos,cosmo_pars,efv,xymap):
    """
    Run the actual fitting procedure.

    :param hdul: an input fits HDUList with extension=0 being the data and extension=1 being the weights.
    :type hdul: list
    :param pos: parameters fit with mcmc
    :type pos: array-like
    :param cosmo_pars: Dictionary of cosmological parameters
    :type cosmo_pars: dict
    :param efv: Dictionary of emcee fitting variables
    :type efv: dict
    :param xymap: tuple of arrays of x- and y-coordinates.
    :type xymap: tuple

    """

    posind = 0
    yint=[]; outalphas=[]
    alp=np.zeros(len(pos))

    if efv["fit_mnlvl"]:
        mnlvl = pos[posind]       
        posind+=1
    else:
        mnlvl = 0
            
    mymap,posind,ynt,myalphas = ellipsoidal_ICM(cosmo_pars,efv,pos,xymap,alp,posind,fixalpha=efv["fixalpha"])
    outmap = mymap
    yint.extend([np.real(ynt)]); outalphas.extend([np.real(myalphas)])

    return outmap,yint,outalphas,mnlvl

def ellipsoidal_ICM(cosmo_pars,efv,pos,xymap,alphas,posind=0,fixalpha=False):
    """
    Run the actual fitting procedure.

    :param cosmo_pars: Dictionary of cosmological parameters
    :type cosmo_pars: dict
    :param efv: Dictionary of emcee fitting variables
    :type efv: dict
    :param pos: emcee fitting parameters
    :type pos: array-like
    :param xymap: tuple of arrays of x- and y-coordinates.
    :type xymap: tuple
    :param alphas: power-law indices
    :type alphas: np.ndarray
    :param posind: index for emcee parameters
    :type posind: int
    :param fixalpha: Fix the power-law indices to input values? Default is False.
    :type fixalpha: bool

    """

    szcv,szcu  = WH.get_sz_values()
    geom       = efv["geo"]
    bins       = efv["bins"]

    nbins = len(bins)
    if efv["finite"] == True:
        nbins-=1     # Important correction!!!
    ulesspres = pos[posind:posind+nbins]
    myalphas = alphas   
    ulessrad  = bins #.to("rad").value
    posind = posind+nbins
    if efv["fit_cen"] == True:
        geom[0:2] = pos[posind:posind+2]  # I think this is what I want...
        posind = posind+2
        
    if efv["fit_geo"] == True:
        geom[2]     = pos[posind]           # Rotation angle
        geom[4]     = pos[posind+1]     # Major axis (should be > 1.0, if minor is defined as 1).
        geom[5]     = np.sqrt(pos[posind+1])     # Assume that the axis along the l.o.s. is geom. mean of maj,min axes.
        pos[posind] = pos[posind] % (np.pi) #if tdtheta > np.pi b/c symmetry reduces it to just pi. (July 2018)
        posind      = posind+2
            
    if efv["model"] == 'NP':
        Int_Pres,outalphas,integrals = ai.integrate_profiles(ulesspres, geom,bins,efv["Thetas"],inalphas=myalphas)

    if efv["model"] == 'GNFW':
        print("Under construction")
        unitless_profile = 0.0
        
    if efv["model"] == 'BETA':
            radii    = efv["Thetas"]
            pprof     = ulesspres[0]*(1.0+(radii/ulesspres[1])**2)**(-1.5*ulesspres[2])        ### Beta model
            #scaling  = scs.gamma(1.5*ulesspres[2] - 0.5)/scs.gamma(1.5*ulesspres[2]) * ulesspres[1] * ulesspres[0]
            scaling  = scs.gamma(1.5*ulesspres[2]-0.5)/scs.gamma(1.5*ulesspres[2])*ulesspres[1] * cosmo_pars["d_a"]
            scaling *= ulesspres[0] * np.sqrt(np.pi) *(szcu['thom_cross'] * u.kpc / szcu['m_e_c2']).to('cm**3 / keV')
            Int_Pres    = scaling.value * (1.0+(radii/ulesspres[1])**2)**(0.5-1.5*ulesspres[2])
            outalphas = Int_Pres*0.0+2.0
            integrals = Int_Pres
            
    #############################################################################################3
        
    if efv["model"] == 'NP':
        pprof,alphas   = ai.log_profile(ulesspres,list(bins),efv["Thetas"]) # Last pos is mn lvl
    elif efv["model"] == 'GNFW':
        pprof = unitless_profile * efv["d_a"]
            
    if efv["ySph"]:
        yint ,newr500=Y_SZ_v2(pprof,efv,retcurvs=False) # As of Aug. 31, 2018
    else:
        yint ,newr500=Y_SZ_v2(Int_Pres,efv,retcurvs=False) # As of Aug. 31, 2018

    #r500,m500,p500, msys = rMP500_from_y500(yinteg,cosmo_pars
    ### Int_Pres natively is in Compton y. If you want a map in different units, one may convert Int_Pres to IntProf
    ### and feed that in instead of Int_Pres below:
    mymap = ai.general_gridding(xymap,efv["Thetas"],bins,geom,finite=efv["finite"],integrals=integrals,
                                            Int_Pres=Int_Pres,oldvs=False)

    return mymap,posind,yint,outalphas

def Y_SZ_v2(yProf,efv,retcurvs=False):
    """
    :param yProf: an integrated profile, i.e. in Compton y that matches theta_range
    :type yProf: np.ndarray
    :param efv: emcee fitting variables
    :type efv: dict
    :param retcurvs: option to return curves; default is False.
    :type retcurvs: bool 

    I'm adopting equations 25-27 in Arnaud+ 2010, which makes use of Y_SZ, or Y_cyl and the
    Universal Pressure Profile (UPP). I tried to find just a straight empirical Y_cyl(R500)-M_500,
    but that doesn't seem to exist?!?

    """

    rProf     = efv["Thetas"]
    r_max     = efv["Theta500"]
    ylist     = efv["y500s"]
    geom      = efv["geo"]
    
    alpha,norm  = ai.ycyl_prep(yProf,rProf)
    
    if not efv["ySph"]:
        yinteg, root,yR,yM = ycyl_simul_v2(rProf,yProf,alpha,r_max,ylist,geom,r_thresh=3e-2,retcurvs=True)
        #import pdb;pdb.set_trace()
    else:
        ### yProf is ***NOT*** a Compton y profile!!! IT IS A PRESSURE PROFILE! -- May 2019
        yinteg, root,yR,yM = ysph_simul(ylist,rProf,yProf,alpha,geom,ythresh=3e-8,retcurvs=True)

    if retcurvs:
        return yinteg, root,yR,yM
    else:
        return yinteg, root

def ycyl_simul_v2(rads,yProf,alpha,maxrad,ylist,geom,r_thresh=3e-2,retcurvs=False):

    """
    :param rads: Radius in radians
    :type rads: numpy.ndarray
    :param yProf: Compton y profile
    :type yProf: numpy.ndarray
    :param alpha: list of power-law indices
    :type alpha: array-like
    :param maxrad: maximum radius
    :type maxrad: float
    :param ylist: array of expected y values.
    :type ylist: numpy.ndarray
    :param geom: [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    :type geom: array-like
    :param r_thresh: Threshold value. Default is 3e-2.
    :type r_thresh: float
    :param retcurvs: Return curves? Default is False.
    :type retcurvs: bool
    """

    fgeo          = geom[3]*geom[4] # Scale by ellipsoidal radii scalings.
    Ycyl          = np.arange(10)
    if alpha[0]  <= -2: alpha[0]=-1.9
    badalp        = (alpha == -2)
    alpha[badalp] = -2.01 # Va fanculo.
    rolledrad     = np.roll(rads,-1)
    intupper      = rolledrad**2 * (rolledrad/rads)**(alpha) #* myrads
    intlower      = rads**2
    intlower[0]   = 0.0
    integrand     = intupper - intlower
    Yshell        = 2.0*np.pi*yProf[:-1]*integrand[:-1]/(alpha[:-1]+2.0)
    Ycyl          = np.cumsum(Yshell)*fgeo
    #import pdb;pdb.set_trace()
    Yref          = ylist[:-1].value
    #Yref          = ylist[:-1]
    mydiff        = Yref - Ycyl

    nrbins        = len(rads)
    #mydiff = rads[1:] - r500
    #absdif = np.abs(mydiff)
    posdiffs = (mydiff > 0)
    turnover = mydiff[posdiffs]
    #import pdb;pdb.set_trace()
    bestr = -rads[1]
    bestY = -Ycyl[0]
    bisca = 0
    if len(turnover) > 1:
        besti  = np.where(mydiff == np.min(turnover))
        bisca  = np.ndarray.item(besti[0])
    if bisca < nrbins -3 and bisca > 10: 
        myinds = bisca + np.asarray([-2,-1,0,1,2],dtype='int')
        #myinds = np.intersect1d(naind,
        myrs   = rads[myinds+1]
        myYs   = Ycyl[myinds]
        myds   = mydiff[myinds]
        myp2   = np.polyfit(myrs,myds,2)
        myY2   = np.polyfit(myrs,myYs,2)
        myroot = np.roots(myp2)
        rdiff  = np.abs(myroot - myrs[2])

        bestr  = myroot[0] if rdiff[0] < rdiff[1] else myroot[1]
        Y2fxn  = np.poly1d(myY2)
        bestY  = Y2fxn(bestr)

        
    #import pdb;pdb.set_trace()
    #bestr  = r500[besti]
    #bestY  = Ycyl[besti]
    if retcurvs:
        return bestY,bestr, Yref,Ycyl
    else:
        return bestY,bestr
   
def ysph_simul(ylist,rads,pProf,alpha,geom,ythresh=3e-8,retcurvs=False):

    """
    :param ylist: array of expected y values.
    :type ylist: numpy.ndarray
    :param rads: Radius in radians
    :type rads: numpy.ndarray
    :param pProf: pressure profile
    :type pProf: numpy.ndarray
    :param alpha: list of power-law indices
    :type alpha: array-like
    :param geom: [X_shift, Y_shift, Rotation, Ella*, Ellb*, Ellc*, Xi*, Opening Angle]
    :type geom: array-like
    :param y_thresh: Threshold value. Default is 3e-8.
    :type y_thresh: float
    :param retcurvs: Return curves? Default is False.
    :type retcurvs: bool
    """  

    fgeo          = geom[3]*geom[4]*geom[5] # Scale by ellipsoidal radii scalings.
    Ysph          = np.arange(10)
    if alpha[0]  <= -3: alpha[0]=-2.9
    badalp        = (alpha == -3)
    alpha[badalp] = -3.01 # Va fanculo.
    rolledrad     = np.roll(rads,-1)
    intupper      = rolledrad**3 * (rolledrad/rads)**(alpha) #* myrads
    intlower      = rads**3
    intlower[0]   = 0.0
    integrand     = intupper - intlower
    Yshell        = 4.0*np.pi*pProf[:-1]*integrand[:-1]/(alpha[:-1]+3.0) 
    Ysph          = np.cumsum(Yshell) *fgeo
    Yref          = ylist[:-1].value
    #Yref          = ylist[:-1]
    mydiff        = Yref - Ysph

    ### Look here (April 13, 2019)
    #import matplotlib.pyplot as plt
    #plt.plot(rads[1:]*3600.0*180/np.pi,Yref)
    #plt.plot(rads[1:]*3600.0*180/np.pi,Ysph)
    #plt.yscale('log');plt.xscale('log')
    #plt.show()
    #import pdb;pdb.set_trace()

    #mydiff = rads[1:] - r500
    #absdif = np.abs(mydiff)
    posdiffs = (mydiff > 0)
    turnover = mydiff[posdiffs]
    #import pdb;pdb.set_trace()
    bestr = rads[51]
    bestY = Ysph[50]
    bisca = 0
    nrbins        = len(rads)
    if len(turnover) > 1:
        besti  = np.where(mydiff == np.min(turnover))
        bisca  = np.ndarray.item(besti[0])
    if bisca < nrbins-3 and bisca > 10: 
        myinds = bisca + np.asarray([-2,-1,0,1,2],dtype='int')
        #myinds = np.intersect1d(naind,
        myrs   = rads[myinds+1]
        myYs   = Ysph[myinds]
        myds   = mydiff[myinds]
        myp2   = np.polyfit(myrs,myds,2)
        myY2   = np.polyfit(myrs,myYs,2)
        myroot = np.roots(myp2)
        rdiff  = np.abs(myroot - myrs[2])

        bestr  = myroot[0] if rdiff[0] < rdiff[1] else myroot[1]
        Y2fxn  = np.poly1d(myY2)
        bestY  = Y2fxn(bestr)

        if bestY > ythresh:
            print(bestr, bestY, np.max(rads))
            stupid = np.random.normal(0,1)
            if stupid > 5: import pdb;pdb.set_trace()
            
        
    #import pdb;pdb.set_trace()
    #bestr  = r500[besti]
    #bestY  = Ysph[besti]
    if retcurvs:
        return bestY,bestr, Yref,Ysph
    else:
        return bestY,bestr

def fit_Gaussian_toSNR(SNRmap,pixsize,maxRad=2.0):
    
    """
    close to fitmap as was used in IDL
    :param SNRmap: a 2D numpy array
    :type SNRmap: nump.ndarray
    :param pixsize: pixel size, in arcseconds
    :type pixsize: float
    :param maxRad: Maximum search radius, arcminutes
    :type maxRad: float
    """

    nx,ny   = SNRmap.shape
    numpyx  = np.outer(np.arange(nx),np.ones(ny))
    numpyy  = np.outer(np.ones(nx),np.arange(ny))
    maxSNR  = np.max(SNRmap)
    maxInd  = (SNRmap == maxSNR)
    #maxinds = np.where(SNRmap == maxSNR)
    #print(maxinds)
    xc      = numpyx[maxInd][0]
    yc      = numpyy[maxInd][0]
    #print(xc,yc)
    #import pdb;pdb.set_trace()
    #xc      = maxinds[0][0]
    #yc      = maxinds[0][1]
    xymap   = WH.get_xymap(SNRmap,pixsize*u.arcsec,xcentre=xc,ycentre=yc,oned=True)
    rmap    = WH.make_rmap(xymap)
    SNRflat = SNRmap.flatten()
    xinit   = np.asarray(xymap)
    grads   = (rmap < maxRad*60.0)
    xdata   = xinit[:,grads]
    ydata   = SNRflat[grads]
    #p0      = np.array([x0,y0,amplitude,sigma,mnlvl])
    xshift  = 1.0  # As a guess
    yshift  = 1.0  # in arcseconds
    sigma   = 20.0 # Gaussian sigma, in arcseconds
    mnlvl   = 1e-2 # Plausible, for an SNR map
    p0      = np.array([xshift,yshift,maxSNR,sigma,mnlvl])
    
    popt, pcov = curve_fit(circ_Gauss, xdata, ydata,p0=p0)

    return popt,pcov,xc,yc
    
def circ_Gauss(xdata,xc,yc,norm,sig,mnlvl):
    """
    Calculate a 2D (circular) Gaussian

    :param xdata: an array of flattened x and y coordinates
    :type xdata: numpy.ndarray
    :param xc: x-center
    :type xc: float
    :param yc: y-center
    :type yc: float
    :param norm: Gaussian amplitude
    :type norm: float
    :param sig: Gaussian width (sigma)
    :type sig: float
    :param mnlvl: mean level (or DC offset, or pedastal)
    :type mnlvl: float
    """

    r1d   = np.sqrt(xdata[0,:]**2 + xdata[1,:]**2)
    Gauss = np.exp(-r1d**2/(2*sig**2))
    model = Gauss*norm + mnlvl
    
    return model

def get_int_SNR(SNRmap,pixsize,maxRad=2.0,bv=120.0,SNRthresh=1.0):
    """
    Get the integrated detection significance.

    :param SNRmap: a 2D numpy array
    :type SNRmap: nump.ndarray
    :param pixsize: pixel size, in arcseconds
    :type pixsize: float
    :param maxRad: Maximum search radius, arcminutes
    :type maxRad: float
    :param bv: beam volume in square arcseconds. Default is 120.
    :type bv: float
    :param SNRthresh: SNR threshold
    :type SNRthresh: float
    """

    popt,pcov,xc,yc        = fit_Gaussian_toSNR(SNRmap,pixsize,maxRad=maxRad)
    xcentre                = xc+popt[0]
    ycentre                = yc+popt[1]
    xymap                  = WH.get_xymap(SNRmap,pixsize*u.arcsec,xcentre=xcentre,ycentre=ycentre,oned=False)
    rmap                   = WH.make_rmap(xymap)
    rbin,ybin,yerr,ycnts   = bin_two2Ds(rmap,SNRmap,binsize=pixsize*2.0)

    badind                 = (ybin < SNRthresh)
    RadThresh              = np.min(rbin[badind])
    goodind                = (rmap < RadThresh)
    goodSNR                = SNRmap[goodind]
    NumberBeams            = np.sum(goodind)*pixsize**2 / bv
    SNRint                 = np.sqrt( np.sum(SNRmap**2)/(NumberBeams) )

    return SNRint,xcentre,ycentre,xymap
    
def bin_two2Ds(independent,dependent,binsize=1,witherr=False,withcnt=False):
    """
    Bins two 2D arrays based on the independent array (e.g. one of radii).

    Parameters
    ----------
    independent : class:`numpy.ndarray`
       An array of independent variables (e.g. radii)
    dependent : class:`numpy.ndarray`
       An array of dependent variables (e.g. RMS or surface brightness)
    binsize : float
       Binsize, relative to independent array.
    witherr : bool
       Calculate the corresponding uncertainties (of the mean)
    withcnt : bool
       Calculate the number of elements (e.g. pixels) within each bin.

    Returns
    -------
    abin : class:`numpy.ndarray`
       Binned absisca values
    obin : class:`numpy.ndarray`
       Binned ordinate values
    oerr : class:`numpy.ndarray`
       Binned uncertainties of the mean
    cnts : class:`numpy.ndarray`
       Binned counts
    """

    flatin = independent.flatten()
    flatnt = dependent.flatten()
    inds = flatin.argsort()

    abscissa = flatin[inds]
    ordinate = flatnt[inds]

    nbins = int(np.ceil((np.max(abscissa) - np.min(abscissa))/binsize))
    abin  = np.zeros(nbins)
    obin  = np.zeros(nbins)
    oerr  = np.zeros(nbins)
    cnts  = np.zeros(nbins) 
    for i in range(nbins):
        blow = i*binsize
        gi = (abscissa >= blow)*(abscissa < blow+binsize)
        abin[i] = np.mean(abscissa[gi])
        obin[i] = np.mean(ordinate[gi])
        if witherr:
            oerr[i] = np.std(ordinate[gi]) / np.sqrt(np.sum(gi))
        if withcnt:
            cnts[i] = np.sum(gi)

    return abin,obin,oerr,cnts

def improved_bin_spacing(radminmax,nbins,mway=False):

    """
    Calculate the bin spacing for MUSTANG-2 based on the number of bins.

    :param radminmax: two-element array or list, of minimum and maximum radii (in arcseconds) to cover.
    :type radminmax: array-like
    :param nbins: Number of bins
    :type nbins: int
    :param mway: An option to try another way. Default (best option) is False.
    :type mway: bool
    """

    m2fwhm    = 10.0 * u.arcsec.to('rad')
    bins      = np.logspace(np.log10(radminmax[0]),np.log10(radminmax[1]), nbins)

    #print(bins)
    
    if mway:
        barr   = bins*1.0
        myinds = np.arange(nbins)
        bsiter = 0
        bspace = barr - np.roll(barr,1)
        bi1    = (bspace < m2fwhm*0.99999)
        bi2    = (myinds > 0)
        badind = [b1 and b2 for b1,b2 in zip(bi1,bi2)]
        while np.sum(badind) > 0:
            minbin = np.min(myinds[badind])
            maxbin = np.max(myinds[badind])
            newbin = (np.arange(np.sum(badind))+1 )*m2fwhm
            stabin = barr[minbin-1]
            barr[badind] = newbin+stabin
            bsiter+=1
            bspace = barr - np.roll(barr,1)
            bi1    = (bspace < m2fwhm*0.99999)
            bi2    = (myinds > 0)
            badind = [b1 and b2 for b1,b2 in zip(bi1,bi2)]
            print("Bin spacing iteration: ",bsiter)

        nlb = nbins-maxbin
        newbin = np.logspace(np.log10(barr[maxbin]),np.log10(radminmax[1]), nlb)
        bins   = barr*1.0
        bins[maxbin:] = newbin
        print("Bins are now: ",barr*3600*180/np.pi,bins*3600*180/np.pi)
        import pdb;pdb.set_trace()

    else:

        print("Bins start with: ",bins*3600*180/np.pi)
        bsiter = 1
        if radminmax[0] == m2fwhm:        
            while bins[1]-bins[0] < m2fwhm:
                #bins   = np.logspace(np.log10(radminmax[0] + m2fwhm*(bsiter-1)),np.log10(radminmax[1]), nbins-bsiter+1)
                bins   = np.logspace(np.log10(m2fwhm*bsiter),np.log10(radminmax[1]), nbins-bsiter+1)
                bsiter+=1
        else:
            while bins[2]-bins[1] < m2fwhm:
                bins   = np.logspace(np.log10(radminmax[0] + m2fwhm*(bsiter-1)),np.log10(radminmax[1]), nbins-bsiter+1)
                bsiter+=1
                
        if bsiter > 2:
            #prebins = np.hstack((np.array([radminmax[0]]),np.arange(1,bsiter-2)*m2fwhm))
            #prebins = np.arange(1,bsiter-1)*m2fwhm
            prebins = np.arange(0,bsiter-2)*m2fwhm + radminmax[0]
            #if radminmax[0] == m2fwhm:
            bins    = np.hstack((prebins,bins))

        print("Bins are now: ",bins*3600*180/np.pi)
    #import pdb;pdb.set_trace()

    return bins

def pos_neg_formatter(med,high_err,low_err,sys=None,cal=None):
    """
    Input the median (or mode), and the *error bars* (not percentile values, but the
    distance between the +/-1 sigma percentiles and the 0 sigma percentile).

    :param med: median
    :type med: float
    :param high_err: uncertainty (on the high side)
    :type high_err: float
    :param low_err: uncertainty (on the low side)
    :type low_err: float
    :param sys: systematic error. Default is None
    :type sys: float
    :param cal: calibration error. Default is None
    :type cal: float
    """

    mypow = np.floor(np.log10(med))
    myexp = 10.0**mypow

    if np.isfinite(mypow):
        if mypow > 0:
            psign = '+'
            pStr = psign+str(int(mypow))
        else:
            pStr = str(int(mypow))
    else:
        # This captures an exceptional case.
        print(med)
        mypow = 0.0
        myexp = 1.0
        pStr = str(int(mypow))
            
    msig  = med/myexp
    hsig  = high_err/myexp
    lsig  = low_err/myexp

    
    msStr = "{:.2F}".format(msig)
    hsStr = '+'+"{:.2F}".format(hsig)
    lsStr = "{:.2F}".format(-lsig)

    baStr = r'${0}^{{{1}}}_{{{2}}}$'. format(msStr,hsStr,lsStr)
    
    if not (sys is None):
        #myma  = (10**sys - 1) * med
        #mymb  = (10**(-sys) - 1) * med
        #import pdb;pdb.set_trace()
        myma  = (np.exp(sys) - 1) * med
        mymb  = (np.exp(-sys) - 1) * med
        hyStr = '+'+"{:.2F}".format(myma/myexp)
        lyStr = "{:.2F}".format(mymb/myexp)
        #hyStr = '+'+"{:.2F}".format(sys/myexp)
        #lyStr = "{:.2F}".format(-sys/myexp)
        baStr = baStr + r' $^{{{0}}}_{{{1}}}$'.format(hyStr,lyStr)
    if not (cal is None):
        hyStr = '+'+"{:.2F}".format(cal[1]/myexp)
        lyStr = "{:.2F}".format(-cal[0]/myexp)
        baStr = baStr + r' $^{{{0}}}_{{{1}}}$'.format(hyStr,lyStr)
    
    exStr = 'E'+pStr
    coStr = baStr+exStr

    return coStr

def extract_radial_profile(hdul,arcmin=True):
    """
    Extract a radial profile using the CRPIX value in a HDU.


    :param hdul: HDUlist
    :type hdul: list
    :param arcmin: If true, then radius is given in arcminutes; otherwise in arcseconds.
    :type arcmin: bool
    """

    SOw                    = WCS(hdul[0].header)
    factor                 = 60.0 if arcmin else 1.0
    pixsize = np.sqrt(np.abs(np.linalg.det(SOw.pixel_scale_matrix))) * 3600.0
    hdulImg                = hdul[0].data

    #print(SOw)
    #import pdb;pdb.set_trace()
    xc                     = SOw.wcs.crpix[0]
    yc                     = SOw.wcs.crpix[1]
    #print(xc,yc)

    xymap                  = WH.get_xymap(hdulImg,pixsize*u.arcsec,xcentre=xc,ycentre=yc,oned=False)
    rmap                   = WH.make_rmap(xymap) / factor
    rbin,ybin,yerr,ycnts   = bin_two2Ds(rmap,hdulImg,binsize=pixsize*2.0/factor)

    return rbin,ybin,yerr,ycnts

def prntPeak(yProf,inHDU,isUK=True,Tag="SimObs-- "):

    """
    Print the peak of a profile and map to STDOUT.

    :param yProf: the (binned) Compton y profile
    :type yProf: numpy.ndarray
    :param inHDU: a HDUList
    :type: inHDU: list
    :param isUK: are the units microKelvin? Default is True
    :type isUK: bool
    :param Tag: Additional tag
    :type Tag: str
    """

    if isUK:
        ProfPeak = np.min(yProf)
        MapPeak  = np.min(inHDU[0].data)
    else:
        ProfPeak = np.max(yProf)
        MapPeak  = np.max(inHDU[0].data)

    print(Tag+"Peak in binned profile: ",ProfPeak," ; peak in map: ",MapPeak)

def plot_SB_profiles(SimObs,SimSky,outdir,filename,prntPk=True,isUK=True,xmin=0,xmax=2.0,
                     addHDU=None,addLabel=None):

    """
    A routine to plot a simulated sky versus simulated observation, radial profiles.

    :param SimObs: an HDUList object for simulated observations
    :type SimObs: list
    :param SimSky: an HDUList object for simulated sky (beam-convolved; not filtered)
    :type SimSky: list
    :param outdir: Output director, with the trailing "/"
    :type outdir: str
    :param filename: Output filename, without path.
    :type filename: str
    :param prntPk: Print the peaks of profiles and maps? Default is True.
    :type prntPk: bool
    :param isUK: Are the arrays in the HDULists in units of uK? Default is True
    :type isUK: bool

    """
    SOr,SOy,SOe,SOc = extract_radial_profile(SimObs)
    SSr,SSy,SSe,SSc = extract_radial_profile(SimSky)
    #WMr,WMy,WMe,WMc = extract_radial_profile(SimSky)
    #pixsize         = WH.get_pixarcsec(SimObs)
    #bmCnt           = WMc*pixsize**2 / bv
    #UncPerBin       = 1.0/ np.sqrt( WMy*bmCnt )

    myfig     = plt.figure(2,figsize=(5,4),dpi=200)
    myfig.clf()
    myax      = myfig.add_subplot(111)
    myax.plot(SOr,SOy,label="Mock Obs")
    myax.plot(SSr,SSy,label="Sky, bm conv.")
    if not addHDU is None:
        addr,addy,adde,addc = extract_radial_profile(addHDU)
        myax.plot(addr,addy,label=addLabel)
        if prntPk:
            prntPeak(addy,addHDU,isUK=True,Tag="Smoothed Mock Obs-- ")

    #myax.plot(WMr,UncPerBin,label="Corresponding Unc.")

    if prntPk:
        prntPeak(SOy,SimObs,isUK=True,Tag="Mock Obs-- ")
        prntPeak(SOy,SimSky,isUK=True,Tag="Bm. Conv. Sky-- ")
    
    myax.set_xlabel("Radius (arcmin)")
    myax.set_ylabel(r"SB ($\mu$K)")
    myax.set_xlim([xmin,xmax])
    myax.legend()

    myfig.tight_layout()
    myfig.savefig(outdir+filename)



import M2_ProposalTools.make_el_chart as mec
import astropy.coordinates as apc
import numpy as np
from astropy import units as u
from datetime import datetime
import os
from astropy.time import Time

from importlib import reload
mec=reload(mec)
### Just some testing lines:
times = ['1999-01-01T00:00:00.123456789', '2010-01-01T00:00:00']
t = Time(times, format='isot', scale='utc')
t.sidereal_time('apparent', 'greenwich')

#########################################################################
### Define the coordinates and name of your object:
#obj_ra = apc.Angle('2h00m00s')
#obj_dec= apc.Angle('-3d00m00s')
### GEMS
#obj_ra = apc.Angle('3h30m00s')
#obj_dec= apc.Angle('-27d49m00s')
#skyobj = apc.SkyCoord(obj_ra, obj_dec, equinox = 'J2000')
#target='GEMS'

#obj_ra = apc.Angle('2h42m46s')
#obj_dec= apc.Angle('-21d32m00s')
#skyobj = apc.SkyCoord(obj_ra, obj_dec, equinox = 'J2000')
#target='PKS0240-217'

#obj_ra = apc.Angle('3h18m00s')
#obj_dec= apc.Angle('41d30m00s')
#skyobj = apc.SkyCoord(obj_ra, obj_dec, equinox = 'J2000')
#target='Perseus'

#obj_ra = apc.Angle('22h42m48s')
#obj_dec= apc.Angle('53d01m00s')
#skyobj = apc.SkyCoord(obj_ra, obj_dec, equinox = 'J2000')
#target='Sausage'

#obj_ra = apc.Angle('02h17m00s')
#obj_dec= apc.Angle('70d30m00s')
#skyobj = apc.SkyCoord(obj_ra, obj_dec, equinox = 'J2000')
#target='CL0217'




#date_obs  = [datetime.strptime('15-01-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-02-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-03-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-04-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-05-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-06-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-07-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-08-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-09-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-10-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-11-2025 23:00:00', '%d-%m-%Y %H:%M:%S'),
#             datetime.strptime('15-12-2025 23:00:00', '%d-%m-%Y %H:%M:%S')]


current_year = datetime.now().year
date_obs  = [datetime.strptime('15-{:02d}-{:d} 23:00:00'.format(mm,current_year), '%d-%m-%Y %H:%M:%S') for mm in np.arange(12,dtype=int)+1]

def get_year_visibilities(elMin=23.0):
    
    obj_ra = apc.Angle('09h30m15.6s')
    obj_dec= apc.Angle('06d15m33.2s')
    skyobj = apc.SkyCoord(obj_ra, obj_dec, equinox = 'J2000')
    target='ACT0930_0615'

    elStr = str(int(elMin))
    mydir = os.getcwd()    # If you want to specify a directory, do so here.

    for i,my_obs_date in enumerate(date_obs):

        monstr = "{:02d}".format(i+1)
        Tag    = "_"+monstr+"Visibility_above"
        mec.plot_visibility(my_obs_date,skyobj,elMin=elMin,mylabel=target,
                            filename = target+Tag+elStr,mydir=mydir)

#mec.plot_contours(elMin=25.0,mylabel="General",filename = "GeneralContourPlot")



import numpy as np
import scipy as sp
from scipy.interpolate import interp1d
import astropy.units as u
import os

def interp_plaws(nodes,norms,alphas,radii):
    """
    Assumes you'll want values extrapolated to min of outradii and max of outradii

    :param nodes: radial nodes
    :type nodes: np.ndarray
    :param norms: normalizations at said nodes
    :type norms: np.ndarray
    :param alphas: the logarithmic slope (power law index) for corresponding nodes.
    :type alphas: np.ndarray
    :param radii: Radii; same units as nodes
    :type radii: np.ndarray
    """
    for node,norm,alpha in zip(nodes,norms,alphas):
        if node == np.min(nodes):
            vals = norm*(radii/node)**alpha
        else:
            gi = (radii > node)
            vals[gi] = norm*(radii[gi]/node)**alpha

    return vals

def int_profile(profrad, profile,radProjected,zmax=0):
    """
    This currently only integrates out to the max of profrad. If you want
    to give a *fixed z*, you should be sure it is LESS THAN the max of
    profrad, and then adjust the code below.

    You likely want profrad to be in kpc. In this way, you will integrate
    units of pressure over kpc, and the resultant units are comprehensible.

    
    :param profrad: radial nodes
    :type profrad: np.ndarray
    :param profile: normalizations at said nodes
    :type profile: np.ndarray
    :param radProjected: is an array of z-coordinates along the line of sight.
    :type radProjected: np.ndarray
    :param zmax: impose a maximum los depth? Default is 0 (no max imposed).
    :type zmax: float, optional.

    """
    
    nrP = len(radProjected); nPr=len(profrad)
    x = np.outer(profrad,np.zeros(nrP)+1.0)
    z = np.outer(np.zeros(nPr)+1.0,radProjected)
    rad = np.sqrt(x**2 + z**2)
    fint = interp1d(profrad, profile, bounds_error = False, fill_value = 0)
    radProfile = fint(rad.reshape(nrP*nPr))
    if zmax > 0:
        zre = z.reshape(nrP*nPr); settozero = (zre > zmax)
        radProfile[settozero] = 0.0
    foo =np.diff(z); bar =foo[:,-1];peloton=radProfile.reshape(nPr,nrP)
    diffz = np.insert(foo,-1,bar,axis=1)
    intProfile = 2.0*np.sum(radProfile.reshape(nPr,nrP)*diffz,axis=1)
    
    return intProfile

def int_tapered_profile_per_theta(profrad, profile,radProjected,mytheta,thetamin=0.4,thetamax=0.6,
                                  zmax=0):
    """
    This currently only integrates out to the max of profrad. If you want
    to give a *fixed z*, you should be sure it is LESS THAN the max of
    profrad, and then adjust the code below.

    NEW EDIT.
    profrad and radProjected should be in RADIANS!!!
    This will make integrating the taper function easier.

    YOU MUST THEREFORE HAVE profile CONVERTED TO THE APPROPRIATE UNITS TO 
    MAKE THIS INTEGRATION CORRECT (scaling/unit-wise).

    :param profrad: radial nodes
    :type profrad: np.ndarray
    :param profile: normalizations at said nodes
    :type profile: np.ndarray
    :param radProjected: is an array of z-coordinates along the line of sight.
    :type radProjected: np.ndarray
    :param mytheta: a given polar angle off of some "nose"
    :type mytheta: float
    :param thetamin: minimum nose angle, wherein a taper begins
    :type thetamin: float
    :param thetamax: maximum nose angle, wherein a taper ends
    :type thetamax: float
    :param zmax: impose a maximum los depth? Default is 0 (no max imposed).
    :type zmax: float, optional.

    """

    #test= 2.0*np.sum(z2,axis=1)
    #thetamin = 0.4
    #thetamax = 0.6
    #profrad  = np.arange(20)
    #profile  = 450.0 - profrad**2
    #radProjected = np.arange(10)

    deltat  = thetamax-thetamin
    
    nrP = len(radProjected); nPr=len(profrad)
    x = np.outer(radProjected,np.ones(nrP))
    y = x * np.tan(mytheta)
    z = np.outer(np.ones(nrP),profrad)
    #print(x.shape,y.shape,z.shape)
    rad     = np.sqrt(x**2 + y**2 + z**2)
    polar   = np.sqrt(z**2 + y**2)
    theta   = np.arctan(polar/x)

    tgttmin = (theta > thetamin)
    tlttmax = (theta < thetamax)
    tgttmax = (theta > thetamax)
    tcos    = tgttmin*tlttmax
    taper   = np.ones(theta.shape)
    taper[tcos] = 0.5 + 0.5*np.cos((theta[tcos]-thetamin)*np.pi/deltat)
    taper[tgttmax] = 0

    radflat = rad.reshape(x.size)
    radProfile = np.interp(radflat,profrad,profile,left=0,right=0)
    #fint = interp1d(profrad, profile, bounds_error = False, fill_value = 0)
    #radProfile = fint(rad.reshape(x.size))
    if zmax > 0:
        zre = z.reshape(nrP*nPr); settozero = (zre > zmax)
        radProfile[settozero] = 0.0
    foo =np.diff(z); bar =foo[:,-1]
    diffz = np.insert(foo,-1,bar,axis=1)
    Pressure3d = radProfile.reshape(taper.shape) * taper
    intProfile = 2.0*np.sum(Pressure3d*diffz,axis=1)
    #print("Why not")
    
    #import pdb; pdb.set_trace()
    
    return intProfile

def loop_int_tppt(thetas,profrad, profile,radProjected,thetamin=0.4,thetamax=0.6,zmax=0):

    """
    Loop over int_tapered_profile_per_theta

    :param thetas: is an array of polar angle off of some "nose"
    :type thetas: np.ndarray
    :param profrad: radial nodes
    :type profrad: np.ndarray
    :param profile: normalizations at said nodes
    :type profile: np.ndarray
    :param radProjected: is an array of z-coordinates along the line of sight.
    :type radProjected: np.ndarray
    :param thetamin: minimum nose angle, wherein a taper begins
    :type thetamin: float
    :param thetamax: maximum nose angle, wherein a taper ends
    :type thetamax: float
    :param zmax: impose a maximum los depth? Default is 0 (no max imposed).
    :type zmax: float, optional.

    """
    
    deltat  = thetamax-thetamin
    
    nrP = len(radProjected); nPr=len(profrad)
    x = np.outer(radProjected,np.ones(nrP))
    z = np.outer(np.ones(nrP),profrad)

    for mytheta in thetas:
        y = x * np.tan(mytheta)
        rad     = np.sqrt(x**2 + y**2 + z**2)
        polar   = np.sqrt(z**2 + y**2)
        theta   = np.arctan(polar/x)
        tgttmin = (theta > thetamin)
        tlttmax = (theta < thetamax)
        tgttmax = np.where(theta > thetamax)
        tcos    = np.where(tgttmin*tlttmax)
        taper   = np.ones(theta.shape)
        taper[tcos] = 0.5 + 0.5*np.cos((theta[tcos]-thetamin)*np.pi/deltat)
        taper[tgttmax] = 0

        radflat = rad.reshape(x.size)
        radProfile = np.interp(radflat,profrad,profile,left=0,right=0)
        if zmax > 0:
            zre = z.reshape(nrP*nPr); settozero = (zre > zmax)
            radProfile[settozero] = 0.0
        foo =np.diff(z); bar =foo[:,-1]
        diffz = np.insert(foo,-1,bar,axis=1)
        Pressure3d = radProfile.reshape(taper.shape) * taper
        intProfile = 2.0*np.sum(Pressure3d*diffz,axis=1)

        if (mytheta == 0):
            Int_Prof = intProfile
        else:
            Int_Prof = np.vstack((Int_Prof,intProfile))

             
    return Int_Prof
        
    
def Ycyl_from_yProf(yprof,rads,r500,nrads=10000):
    """
    Calculate y_cylindrival from profile

    :param yprof: Compton y profile
    :type yprof: np.ndarray
    :param rads: radii
    :type rads: np.ndarray
    :param r500: :math:`R_{500}`
    :type r500: float
    :param nrads: Number of radii for use with interpolation
    :type nrads: int
    """
    
    radVals = np.arange(nrads)*r500/nrads
    drad    = r500/nrads
    yVals   = np.interp(radVals, rads,yprof)
    goodr   = (radVals < r500)
    
    Yint = np.sum(yVals[goodr] * 2.0 * np.pi * radVals[goodr] * drad)

    return Yint
